from typing import List, Optional

import pydantic


class ModelParams(pydantic.BaseModel):
    loss: List[int] = pydantic.Field(
        description="List of ints signifying loss per asset"
    )
    min_loss: Optional[int] = pydantic.Field(
        description="Minimum possible loss for the model "
    )


class GaussianModelInput(ModelParams):
    num_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of qubits represent"
        "the latent normal random variable Z (Resolution of "
        "the random variable Z)."
    )
    normal_max_value: float = pydantic.Field(
        description="Min/max value to truncate the " "latent normal random variable Z"
    )
    default_probabilities: List[pydantic.confloat(ge=0, le=1)] = pydantic.Field(
        description="default probabilities for each asset"
    )

    rhos: List[pydantic.confloat(gt=0)] = pydantic.Field(
        description="Sensitivities of default probability of assets "
        "with respect to Z (1/sigma(Z))"
    )

    @property
    def num_model_qubits(self):
        return len(self.rhos)



import enum
from datetime import time, datetime
from typing import Union, Dict, Tuple, List, Optional

import pydantic

Plot = pydantic.conint(ge=0)
Track = pydantic.conint(ge=0)
MHTSolution = Dict[Plot, Track]


# We added this class to avoid issues with the schema, as it doesn't handle List[Tuple[int, int] or a union of
# List[List[int] with other things
class MhtSegmentSolution(pydantic.BaseModel):
    segment_list: List[Tuple[int, ...]]

    @pydantic.validator("segment_list")
    def verify_segments_are_legal(cls, segment_list):
        if any(len(segment) != 2 for segment in segment_list):
            raise ValueError("Segment list contains a segment with more than two plots")

        return segment_list


Solution = Union[Tuple[int, ...], MhtSegmentSolution, MHTSolution]


class HybridStatus(str, enum.Enum):
    SUCCESS = "success"
    ERROR = "error"


class SolverResult(pydantic.BaseModel):
    best_cost: float
    # TODO: add time units (like seconds)
    time: time
    solution: Solution


class SolutionData(pydantic.BaseModel):
    solution: Solution
    repetitions: pydantic.PositiveInt
    cost: float


class VQEIntermediateData(pydantic.BaseModel):
    utc_time: datetime = pydantic.Field(description="Time when the iteration finished")
    iteration_number: pydantic.PositiveInt = pydantic.Field(
        description="The iteration's number (evaluation count)"
    )
    parameters: List[float] = pydantic.Field(
        description="The optimizer parameters for the variational form"
    )
    mean_all_solutions: Optional[float] = pydantic.Field(
        default=None, description="The mean score of all solutions in this iteration"
    )
    mean_valid_solutions: Optional[float] = pydantic.Field(
        default=None, description="The mean score of the valid solutions."
    )
    solutions: List[SolutionData] = pydantic.Field(
        description="Solutions found in this iteration, their score and"
        "number of repetitions"
    )


class VQASolverResult(SolverResult):
    solution_distribution: Optional[List[SolutionData]]
    intermediate_results: Optional[List[VQEIntermediateData]]


class VQESolverResult(VQASolverResult):
    pass


class QAOASolverResult(VQASolverResult):
    preprocessed_graph: Dict
    solution_segment: List[Tuple[int, ...]]


class SolverResults(pydantic.BaseModel):
    vqe: Union[
        QAOASolverResult, VQESolverResult
    ]  # From most specific, to least specific
    classical: Optional[SolverResult]


class HybridResult(pydantic.BaseModel):
    status: HybridStatus
    details: Union[SolverResults, str]



import enum
from classiq_interface.generator.problem_name_mapping import PROBLEM_INPUT_MAPPING
from classiq_interface.hybrid.mht_input import MHTInput
from classiq_interface.hybrid.mht_qaoa_input import MhtQaoaInput
from classiq_interface.hybrid.mvc_input import MVCInput
from classiq_interface.hybrid.tsp_input import TSPInput


class MixerType(enum.Enum):
    pass


class TSPMixerType(MixerType):
    CY_CUSTOM = "CY_CUSTOM"  # CY gates, with a target qubit that is part of a register encoding '11..1' domain range,
    # and a control qubit from another register
    CY_CUSTOM_INNER = "CY_CUSTOM_INNER"  # CY gates, with a target qubit that is part of a register encoding '11..1' domain
    # range, and a control qubit from the same register
    CZ = "CZ"  # CZ gates between each consecutive qubits pair
    CZ_INNER = "CZ_INNER"  # CZ gates between qubits in same register
    CY_INNER = "CY_INNER"  # CY gates between qubits in same register


MIXER_TYPES_DEFAULT = {
    TSPInput: [TSPMixerType.CY_CUSTOM, TSPMixerType.CZ],
    MVCInput: [],
    MHTInput: [],
    MhtQaoaInput: [],
}

for e in PROBLEM_INPUT_MAPPING.values():
    assert e in MIXER_TYPES_DEFAULT, "verify all problem inputs are accessible"



import ast
import enum
import keyword
import math
import random
import re
from enum import Enum
from typing import Union, Optional, Generic, TypeVar, Dict, List

import pydantic
from classiq_interface.generator.function_params import FunctionParams
from pydantic.generics import GenericModel

_MAX_FRACTION_PLACES = 5
DEFAULT_ARG_NAME = "arg"
DEFAULT_RIGHT_ARG_NAME = "right_arg"
DEFAULT_LEFT_ARG_NAME = "left_arg"
DEFAULT_OUT_NAME = "out"
LeftDataT = TypeVar("LeftDataT")
RightDataT = TypeVar("RightDataT")
Numeric = (float, int)


class FixPointNumber(pydantic.BaseModel):
    float_value: float
    max_fraction_places: pydantic.conint(ge=0) = _MAX_FRACTION_PLACES
    is_signed: Optional[bool] = None
    _fraction_places: Optional[int] = pydantic.PrivateAttr(default=None)
    _int_val: Optional[int] = pydantic.PrivateAttr(default=None)
    _size: Optional[int] = pydantic.PrivateAttr(default=None)
    _integer_part_size: Optional[int] = pydantic.PrivateAttr(default=None)

    def set_int_representation(self):
        int_val = math.floor(self.float_value * 2 ** self.max_fraction_places)
        int_val = self.signed_int_to_unsigned(int_val)

        if int_val == 0:
            fraction_places = 0
        else:
            bin_val = bin(int_val)[2:]
            fraction_places = self.max_fraction_places
            for b in reversed(bin_val):
                if b == "1" or fraction_places == 0:
                    break
                fraction_places -= 1
            int_val = int_val >> (self.max_fraction_places - fraction_places)

        self._fraction_places = fraction_places
        self._int_val = int_val

    @staticmethod
    def signed_int_to_unsigned(number: int):
        """Return the integer value of a signed int if it would we read as un-signed in binary representation"""
        if number >= 0:
            return number

        not_power2 = abs(number) & (abs(number) - 1) != 0
        return number + 2 ** (number.bit_length() + 1 * not_power2)

    @staticmethod
    def binary_to_float(bin_rep: str, fraction_part_size: int, is_signed: bool = False):
        negative_offset = -(2 ** len(bin_rep)) * (bin_rep[0] == "1") * is_signed
        value = int(bin_rep, 2) + negative_offset
        if (
            fraction_part_size > 0
        ):  # this separated to make sure that the value remains int if no there is no fraction part
            value = value / 2 ** fraction_part_size
        return value

    @pydantic.validator("is_signed", always=True)
    def validate_is_signed(cls, is_signed: bool, values: dict):
        float_value = values.get("float_value")
        if is_signed is False and float_value < 0:
            raise ValueError("Not possible to define a negative number as not signed")
        elif is_signed is None:
            is_signed = float_value < 0

        return is_signed

    @property
    def fraction_places(self):
        if self._fraction_places is None:
            self.set_int_representation()
        return self._fraction_places

    def set_fraction_places(self, value: int):
        if value > self.max_fraction_places:
            self.max_fraction_places = value
            self.set_int_representation()

        self._int_val = math.floor(self.int_val * 2 ** (value - self.fraction_places))
        self._fraction_places = value
        self._size = self._integer_part_size + self._fraction_places

    @property
    def int_val(self):
        if self._int_val is None:
            self.set_int_representation()
        return self._int_val

    @property
    def integer_part_size(self):
        if self._integer_part_size is None:
            self._integer_part_size = self.int_val.bit_length() - self.fraction_places
        return self._integer_part_size

    def set_integer_part_size(self, value: int):
        if value < self.integer_part_size:
            raise ValueError("size cannot be lower than minimum number bits required")
        self._integer_part_size = value
        self._size = self._integer_part_size + self._fraction_places
        self._int_val = int(self.bin_val, 2)

    def bit_length(self):
        return 1 if self.int_val == 0 else self.int_val.bit_length()

    @property
    def size(self):
        if self._size is None:
            self._size = self.bit_length()
        return self._size

    @property
    def bin_val(self):
        if self._int_val is None:
            self.set_int_representation()

        bin_rep = bin(self._int_val)[2:]
        size_diff = self.size - len(bin_rep)
        if self.float_value >= 0:
            return "0" * size_diff + bin_rep
        else:
            return "1" * size_diff + bin_rep

    @property
    def actual_float_value(self):
        return self.binary_to_float(self.bin_val, self.fraction_places, self.is_signed)

    def __getitem__(self, item):
        return [v for v in self.bin_val[::-1]][
            item
        ]  # follow qiskit convention that LSB is the top wire, bigendian

    class Config:
        extra = "forbid"


class RegisterUserInput(pydantic.BaseModel):
    size: pydantic.PositiveInt
    name: Optional[str] = None
    is_signed: bool = False
    fraction_places: pydantic.conint(ge=0) = 0

    class Config:
        extra = "forbid"


class UnaryOpParams(FunctionParams):
    arg: RegisterUserInput
    output_size: Optional[pydantic.PositiveInt]
    output_name: Optional[str]
    inplace: bool = False

    def create_io_enums(self):
        output_name = self.output_name if self.output_name else DEFAULT_OUT_NAME
        self._output_names = Enum("BinaryOpOutputs", {output_name: output_name})

        arg_name = self.arg.name if self.arg.name else DEFAULT_ARG_NAME

        self._input_names = Enum(
            "UnaryOpInputs",
            {arg_name: arg_name},
        )

    class Config:
        arbitrary_types_allowed = True


class LogicalAnd(FunctionParams):
    args: List[Union[RegisterUserInput, FixPointNumber, int, float]]
    output_name: Optional[str] = DEFAULT_OUT_NAME

    @pydantic.validator("args")
    def validate_inputs_sizes(cls, args):
        for arg in args:
            if (
                isinstance(arg, RegisterUserInput)
                and arg.size != 1
                and arg.fraction_places != 0
            ):
                raise ValueError(
                    f"All inputs to logical and must be of size 1 | {arg.name}"
                )

        return args

    @pydantic.validator("args")
    def set_inputs_names(cls, args):
        for i, arg in enumerate(args):
            if isinstance(arg, RegisterUserInput):
                arg.name = arg.name if arg.name else DEFAULT_ARG_NAME + str(i)

        return args

    def create_io_enums(self):
        output_name = self.output_name
        self._output_names = Enum("BinaryOpOutputs", {output_name: output_name})

        arg_names = [
            arg.name for arg in self.args if isinstance(arg, RegisterUserInput)
        ]

        self._input_names = Enum(
            "LogicalAndInputs",
            {arg_name: arg_name for arg_name in arg_names},
        )

    class Config:
        arbitrary_types_allowed = True


class BitwiseInvert(UnaryOpParams):
    pass


class Negation(UnaryOpParams):
    pass


class BinaryOpParams(GenericModel, FunctionParams, Generic[LeftDataT, RightDataT]):
    left_arg: Union[LeftDataT, RegisterUserInput]
    right_arg: Union[RightDataT, RegisterUserInput]
    output_size: Optional[pydantic.PositiveInt]
    output_name: Optional[str]

    @pydantic.root_validator()
    def validate_one_is_register(cls, values):
        if isinstance(values.get("left_arg"), Numeric) and isinstance(
            values.get("right_arg"), Numeric
        ):
            raise ValueError("One argument must be a register")
        return values

    def create_io_enums(self):
        output_name = self.output_name if self.output_name else DEFAULT_OUT_NAME
        self._output_names = Enum("BinaryOpOutputs", {output_name: output_name})

        if isinstance(self.left_arg, RegisterUserInput) and isinstance(
            self.right_arg, RegisterUserInput
        ):
            left_arg_name = (
                self.left_arg.name if self.left_arg.name else DEFAULT_LEFT_ARG_NAME
            )
            right_arg_name = (
                self.right_arg.name if self.right_arg.name else DEFAULT_RIGHT_ARG_NAME
            )
            self._input_names = Enum(
                "BinaryOpInputs",
                {left_arg_name: left_arg_name, right_arg_name: right_arg_name},
            )
            return

        if isinstance(self.left_arg, RegisterUserInput):
            arg_name = (
                self.left_arg.name if self.left_arg.name else DEFAULT_LEFT_ARG_NAME
            )
        else:
            assert isinstance(
                self.right_arg, RegisterUserInput
            ), "At least one argument should be a register"
            arg_name = (
                self.right_arg.name if self.right_arg.name else DEFAULT_RIGHT_ARG_NAME
            )

        self._input_names = Enum("BinaryOpInputs", {arg_name: arg_name})

    class Config:
        arbitrary_types_allowed = True


class BinaryOpWithIntInputs(BinaryOpParams[int, int]):
    @pydantic.root_validator()
    def validate_int_registers(cls, values):
        left_arg = values.get("left_arg")
        is_left_arg_float_register = (
            isinstance(left_arg, RegisterUserInput) and left_arg.fraction_places > 0
        )
        right_arg = values.get("right_arg")
        is_right_arg_float_register = (
            isinstance(right_arg, RegisterUserInput) and right_arg.fraction_places > 0
        )
        if is_left_arg_float_register or is_right_arg_float_register:
            raise ValueError("Boolean operation are defined only for integer")

        return values


class BinaryOpWithFloatInputs(
    BinaryOpParams[Union[float, FixPointNumber], Union[float, FixPointNumber]]
):
    @pydantic.validator("left_arg", "right_arg")
    def convert_numeric_to_fix_point_number(cls, val):
        if isinstance(val, Numeric):
            val = FixPointNumber(float_value=val)
        return val


class BitwiseAnd(BinaryOpWithIntInputs):
    pass


class BitwiseOr(BinaryOpWithIntInputs):
    pass


class BitwiseXor(BinaryOpWithIntInputs):
    pass


class Adder(BinaryOpWithFloatInputs):
    inplace: bool = True


class Equal(BinaryOpWithFloatInputs):
    @pydantic.validator("output_size")
    def validate_output_size(cls, val):
        if val and val != 1:
            raise ValueError("Equal function has only one output qubits")
        return val


class MappingMethods(str, enum.Enum):
    topological = "topological"
    pebble = "pebble"


class Arithmetic(FunctionParams):
    max_fraction_places: pydantic.conint(ge=0) = _MAX_FRACTION_PLACES
    expression: str
    definitions: Dict[
        str,
        Union[
            pydantic.StrictInt, pydantic.StrictFloat, FixPointNumber, RegisterUserInput
        ],
    ]
    method: MappingMethods = MappingMethods.pebble
    qubit_count: Optional[pydantic.conint(ge=0)] = None
    output_name: Optional[str]
    random_seed: Optional[int] = None

    @pydantic.validator("expression")
    def check_expression_is_legal(cls, expression):
        ast.parse(expression, "", "eval")
        return expression

    @pydantic.root_validator()
    def check_all_variable_are_defined(cls, values):
        expression, definitions = values.get("expression"), values.get("definitions")

        literals = re.findall("[A-Za-z][A-Za-z0-9]*", expression)
        undefined_literals = {
            literal for literal in literals if literal not in definitions
        }.difference(set(keyword.kwlist))

        if undefined_literals:
            raise ValueError(f"{undefined_literals} need to be defined in definitions")
        return values

    @pydantic.root_validator()
    def substitute_expression(cls, values):
        # TODO there isn't a secure way to simplify the expression which does not involve using eval.
        #  Can be done with sdk on client side
        expression, definitions = values.get("expression"), values.get("definitions")
        new_definition = dict()
        for var_name, value in definitions.items():
            if isinstance(value, RegisterUserInput):
                new_definition[var_name] = value
                continue
            elif isinstance(value, int):
                pass
            elif isinstance(value, float):
                value = FixPointNumber(float_value=value).actual_float_value
            elif isinstance(value, FixPointNumber):
                value = value.actual_float_value
            else:
                raise ValueError(f"{type(value)} type is illegal")

            expression = re.sub(r"\b" + var_name + r"\b", str(value), expression)
        values["expression"] = expression
        values["definitions"] = new_definition
        return values

    @pydantic.validator("definitions")
    def set_register_names(cls, definitions):
        for k, v in definitions.items():
            if isinstance(v, RegisterUserInput):
                v.name = k
        return definitions

    @pydantic.validator("random_seed", always=True)
    def validate_random_seed(cls, random_seed):
        return random.randint(0, 1000) if not random_seed else random_seed

    def create_io_enums(self):
        literals = re.findall("[A-Za-z][A-Za-z0-9]*", self.expression)
        output_name = self.output_name if self.output_name else DEFAULT_OUT_NAME
        self._input_names = Enum(
            "ArithmeticInputs",
            {literal: literal for literal in literals if literal not in keyword.kwlist},
        )
        self._output_names = Enum("ArithmeticOutputs", {output_name: output_name})

    class Config:
        extra = "forbid"


class ArithmeticOracle(Arithmetic):
    @pydantic.validator("expression")
    def validate_compare_expression(cls, value):
        ast_obj = ast.parse(value, "", "eval")
        if not isinstance(ast_obj.body, (ast.Compare, ast.BoolOp)):
            raise ValueError("Must a comparison expression")

        return value



import pydantic
from classiq_interface.generator.circuit_outline import Qubit, Cycle


class SingleRegisterBarrier(pydantic.BaseModel):
    cycle: Cycle


class SingleQubitBarrier(pydantic.BaseModel):
    qubit: Qubit
    cycle: Cycle



from typing import Optional, Tuple, Generic, TypeVar

import pydantic
from pydantic.generics import GenericModel

RangeType = TypeVar("RangeType", int, float)


class Range(GenericModel, Generic[RangeType]):
    lower_bound: Optional[RangeType] = None
    upper_bound: Optional[RangeType] = None

    @pydantic.validator("upper_bound")
    def validate_bounds_order(cls, upper_bound, values):
        lower_bound = values.get("lower_bound")

        if (
            lower_bound is not None
            and upper_bound is not None
            and lower_bound > upper_bound
        ):
            raise ValueError("lower bound must not be greater than upper bound")

        return upper_bound

    # TODO: Remove default values
    def to_tuple(
        self,
        default_lower_bound: RangeType = -1,
        default_upper_bound: RangeType = float("inf"),
    ) -> Tuple[RangeType, RangeType]:
        return (
            self.lower_bound if self.lower_bound else default_lower_bound,
            self.upper_bound if self.upper_bound else default_upper_bound,
        )


NonNegativeIntRange = Range[pydantic.conint(ge=0)]
NonNegativeFloatRange = Range[pydantic.confloat(ge=0)]



from typing import List, Optional

import pydantic

from classiq_interface.generator.range_types import NonNegativeIntRange

# TODO figure how to create a pydantic class that can serve as dictionary key
# TODO openapi raises error when using pydantic.constr, need to find solution
# check against duplicates and give better errors
ValidGateName = str


class CustomGateImplementation(pydantic.BaseModel):
    # TODO: add validation for qasm string?
    qasm: str
    n_qubits: pydantic.PositiveInt
    name: Optional[ValidGateName] = None
    depth: Optional[pydantic.PositiveInt] = None


class CustomGate(NonNegativeIntRange):
    implementations: List[CustomGateImplementation]
    add_as_single_gate: bool = True



import numpy as np
import pydantic


class LogNormalModelInput(pydantic.BaseModel):
    num_qubits: pydantic.PositiveInt = pydantic.Field(
        description="Number of qubits to represent the probability."
    )
    mu: pydantic.confloat(ge=0) = pydantic.Field(
        description="Mean of the Normal distribution variable X s.t. ln(X) ~ log-normal."
    )
    sigma: pydantic.confloat(gt=0) = pydantic.Field(
        description="Std of the Normal distribution variable X s.t. ln(X) ~ log-normal."
    )

    @property
    def bounds(self):
        mean = np.exp(self.mu + self.sigma ** 2 / 2)
        variance = (np.exp(self.sigma ** 2) - 1) * np.exp(2 * self.mu + self.sigma ** 2)
        stddev = np.sqrt(variance)
        low = np.maximum(0, mean - 3 * stddev)
        high = mean + 3 * stddev
        return low, high

    @property
    def num_model_qubits(self):
        return self.num_qubits



import secrets
from typing import Tuple


def random_probability(options_count: int) -> Tuple[float, ...]:
    random_generator = secrets.SystemRandom()
    random_numbers = tuple(
        random_generator.uniform(0, 10) for _ in range(options_count)
    )

    total = sum(random_numbers)
    assert total != 0, "Error creating random probability."

    return tuple(i / total for i in random_numbers)



import pydantic
from pydantic import json


# Use this class as a type for complex data from the json, e.g., in the state_propagator function.
class Complex(complex):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, str):
            v = "".join(v.split())

        return complex(v)


assert (
    complex not in pydantic.json.ENCODERS_BY_TYPE
), "Is complex type supported on newer version of pydantic?"
pydantic.json.ENCODERS_BY_TYPE[complex] = str



from enum import Enum
from typing import Type, List
import numpy as np

import pydantic

from classiq_interface.generator.complex_type import Complex
from classiq_interface.generator.function_params import FunctionParams
from classiq_interface.generator.sparse_amp_load import amplitudes_sum_to_one
from classiq_interface.generator.state_preparation import is_power_of_two


class HadamardAmpLoadOutputs(Enum):
    OUTPUT_STATE = "OUTPUT_STATE"


class HadamardAmpLoad(FunctionParams):
    """
    loads a amplitudes vector using hadamard decomposition
    """

    num_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of qubits in the circuit."
    )
    amplitudes: List[Complex] = pydantic.Field(description="amplitudes vector to load")

    cutoff: pydantic.conint(ge=1) = pydantic.Field(
        description="The number of hadamard coefficients to keep. "
        "The largest cutoff_num coefficients are used to load the amplitudes"
    )

    _is_power_of_two = pydantic.validator("amplitudes", allow_reuse=True)(
        is_power_of_two
    )
    _is_sum_to_one = pydantic.validator("amplitudes", allow_reuse=True)(
        amplitudes_sum_to_one
    )

    @pydantic.root_validator()
    def cutoff_validator(cls, values):
        amp = values.get("amplitudes")
        cutoff = values.get("cutoff")
        if cutoff > len(amp):
            raise ValueError(
                "cutoff number should be smaller or equal to the length of the amplitudes vector "
            )
        return values

    _output_names: Type[Enum] = pydantic.PrivateAttr(default=HadamardAmpLoadOutputs)



import math
import networkx as nx
from typing import Tuple, Dict

from pyomo.core import Objective, minimize, Constraint
import pyomo.environ as pyo


Node = int
Edge = Tuple[Node, Node]
Pubo = Dict[Tuple[Edge, ...], float]


def build_mht_pyomo_model(
    pubo: Pubo, scenario_graph: nx.DiGraph, has_constraints: bool = True
) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()
    model.Nodes = pyo.Set(initialize=list(scenario_graph.nodes))
    model.Arcs = pyo.Set(initialize=list(scenario_graph.edges))
    model.x = pyo.Var(model.Arcs, domain=pyo.Binary)

    _DECIMALS = 3

    if has_constraints:

        @model.Constraint(model.Nodes)
        def out_edges_rule(model, idx):
            out_nodes = [
                node_id for node_id in model.Nodes if [idx, node_id] in model.Arcs
            ]
            if len(out_nodes) >= 2:
                return sum(model.x[idx, node_id] for node_id in out_nodes) <= 1
            else:
                return Constraint.Feasible

        @model.Constraint(model.Nodes)
        def in_edges_rule(model, idx):
            in_nodes = [
                node_id for node_id in model.Nodes if [node_id, idx] in model.Arcs
            ]
            if len(in_nodes) >= 2:
                return sum(model.x[node_id, idx] for node_id in in_nodes) <= 1
            else:
                return Constraint.Feasible

    def obj_expression(model):
        return sum(
            round(pubo_energy, _DECIMALS)
            * math.prod(model.x[edge] for edge in pubo_edges)
            for pubo_edges, pubo_energy in pubo.items()
        )

    model.cost = Objective(rule=obj_expression, sense=minimize)

    return model



import base64
import enum
import io
from typing import Union, List, Optional

import pydantic
from PIL import Image
from classiq_interface.generator.generation_metadata import GenerationMetadata


class QuantumFormat(str, enum.Enum):
    QASM = "qasm"
    QSHARP = "qs"


class GenerationStatus(str, enum.Enum):
    NONE = "none"
    SUCCESS = "success"
    UNSAT = "unsat"
    TIMEOUT = "timeout"
    CANCELLED = "cancelled"
    ERROR = "error"


class GeneratedCircuit(pydantic.BaseModel):
    qasm: Optional[str]
    qsharp: Optional[str]
    output_format: List[QuantumFormat]
    image: str
    metadata: GenerationMetadata

    def show(self) -> None:
        image = Image.open(io.BytesIO(base64.b64decode(self.image)))
        image.show()


class GenerationResult(pydantic.BaseModel):
    status: GenerationStatus
    details: Union[GeneratedCircuit, str]



import enum

import pydantic


class OptimizationProblemName(str, enum.Enum):
    TSP = "TSP"
    MVC = "MVC"
    MHT = "MHT"
    MhtQaoa = "MHT QAOA"


class ProblemInput(pydantic.BaseModel):
    name: OptimizationProblemName

    class Config:
        # Using this class allows sending this object as a json to the
        # frontEnd with all the objects that inheritance from this class
        extra = "allow"

    def is_valid_cost(self, cost: float) -> bool:
        raise NotImplementedError



import enum
from typing import Union

import pydantic
from classiq_interface.finance.gaussian_model_input import GaussianModelInput
from classiq_interface.finance.log_normal_model_input import LogNormalModelInput

Models = Union[GaussianModelInput, LogNormalModelInput]


class FinanceModelName(str, enum.Enum):
    GAUSSIAN = "gaussian"
    LOG_NORMAL = "log normal"


class FinanceModelInput(pydantic.BaseModel):
    name: Union[FinanceModelName, str]
    params: Models



from typing import Union

from classiq_interface.hybrid.encoding_type import EncodingType
from classiq_interface.hybrid.mht_input import MHTInput
from classiq_interface.hybrid.mht_qaoa_input import MhtQaoaInput
from classiq_interface.hybrid.mvc_input import MVCInput
from classiq_interface.hybrid.problem_input import OptimizationProblemName, ProblemInput
from classiq_interface.hybrid.tsp_input import TSPInput

PROBLEM_INPUT_MAPPING = {
    OptimizationProblemName.TSP: TSPInput,
    OptimizationProblemName.MVC: MVCInput,
    OptimizationProblemName.MHT: MHTInput,
    OptimizationProblemName.MhtQaoa: MhtQaoaInput,
}

ENCODING_TYPE_DEFAULT = {
    TSPInput: EncodingType.NODES,
    MVCInput: EncodingType.DICKE,
    MHTInput: EncodingType.SERIAL,
    MhtQaoaInput: EncodingType.SERIAL,
}

for e in OptimizationProblemName:
    assert e in PROBLEM_INPUT_MAPPING, "verify all problem inputs are accessible"

for e in PROBLEM_INPUT_MAPPING.values():
    assert e in ENCODING_TYPE_DEFAULT, "verify all problem inputs are accessible"


def parse_problem_input(problem: Union[dict, ProblemInput]) -> ProblemInput:
    problem_name = (
        problem.name if isinstance(problem, ProblemInput) else problem["name"]
    )
    return PROBLEM_INPUT_MAPPING[problem_name].parse_obj(obj=problem)



import random


def create_random_seed():
    _MIN_SEED = 0
    _MAX_SEED = (
        2 ** 32 - 1
    )  # not certain of underline limitation in minizinc, 32bit should enough
    return random.SystemRandom().randint(a=_MIN_SEED, b=_MAX_SEED)



from typing import Optional, List, Union

import pydantic

from classiq_interface.generator.generation_metadata import GenerationMetadata
from classiq_interface.backend.backend_preferences import (
    AwsBackendPreferences,
    IBMBackendPreferences,
    AzureBackendPreferences,
)


class AmplitudeEstimation(pydantic.BaseModel):
    alpha: float = pydantic.Field(
        default=0.05, description="Confidence level of the AE algorithm"
    )
    epsilon: float = pydantic.Field(
        default=0.01, description="precision for estimation target `a`"
    )
    binary_search_threshold: Optional[pydantic.confloat(ge=0, le=1)] = pydantic.Field(
        description="The required probability on the tail of the distribution (1 - percentile)"
    )


class AmplitudeAmplification(pydantic.BaseModel):
    iterations: Union[List[int], int, None] = pydantic.Field(
        default=None, description="Number or list of numbers of iteration to use"
    )
    growth_rate: Optional[float] = pydantic.Field(
        default=None,
        description="Number of iteration used is set to round(growth_rate**iterations)",
    )
    sample_from_iterations: bool = pydantic.Field(
        default=False,
        description="If True, number of iterations used is picked randomly from [1, iteration] range",
    )


class GeneralPreferences(pydantic.BaseModel):
    num_shots: int = 100
    amplitude_estimation: Optional[AmplitudeEstimation] = pydantic.Field(
        default_factory=AmplitudeEstimation
    )
    amplitude_amplification: Optional[AmplitudeAmplification] = pydantic.Field(
        default_factory=AmplitudeAmplification
    )
    backend_preferences: Union[
        AzureBackendPreferences, IBMBackendPreferences, AwsBackendPreferences
    ] = pydantic.Field(
        default_factory=lambda: IBMBackendPreferences(
            backend_service_provider="IBMQ", backend_name="aer_simulator_statevector"
        ),
        description="Preferences for the requested backend to run the quantum circuit.",
    )


class ExecuteProblem(pydantic.BaseModel):
    problem_preferences: GeneralPreferences = pydantic.Field(
        default_factory=GeneralPreferences, description="preferences for the execution"
    )

    problem_data: GenerationMetadata = pydantic.Field(
        default=None, description="Data returned from the generation procedure."
    )



from enum import Enum
from typing import Type

import pydantic
from classiq_interface.generator.function_params import FunctionParams


class IntegerComparatorInputs(Enum):
    ARGUMENT_IN = "ARGUMENT_IN"
    TARGET_QUBIT = "TARGET_QUBIT"


class IntegerComparatorOutputs(Enum):
    ARGUMENT_OUT = "ARGUMENT_OUT"
    COMPARISON_RESULT = "COMPARISON_RESULT"


class IntegerComparator(FunctionParams):
    """
    Compares basis states |i> against a given integer n.
     if greater_or_equal = True, the target qubit is flipped if i is greater or equal to n
     if greater_or_equal = False, the target qubit is flipped if i is smaller than n
    """

    num_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of qubits describing the input state."
    )
    comparand: int = pydantic.Field(description="integer to compare with")
    greater_or_equal: bool = pydantic.Field(
        default=True,
        description="if True, evaluates if input state is >= than given integer,"
        " otherwise <",
    )

    _input_names: Type[Enum] = pydantic.PrivateAttr(default=IntegerComparatorInputs)
    _output_names: Type[Enum] = pydantic.PrivateAttr(default=IntegerComparatorOutputs)



from typing import Literal, List

import more_itertools
import pydantic
import numpy as np

from classiq_interface.hybrid.problem_input import ProblemInput, OptimizationProblemName

_TOLERANCE_DECIMALS = 6


class PlotData(pydantic.BaseModel):
    # We are currently ignoring units. This might need to be handled in the future
    x: float = pydantic.Field(description="The X coordinate of this plot")
    y: float = pydantic.Field(description="The Y coordinate of this plot")
    t: float = pydantic.Field(description="The time stamp of this plot")
    plot_id: pydantic.conint(ge=0) = pydantic.Field(
        description="The plot ID of this plot"
    )


class MhtQaoaInput(ProblemInput):
    name: Literal[OptimizationProblemName.MhtQaoa] = pydantic.Field(
        default=OptimizationProblemName.MhtQaoa,
        description="Name of optimization problem.",
    )
    reps: pydantic.conint(ge=1) = pydantic.Field(
        default=3, description="Number of QAOA layers."
    )
    plot_list: List[PlotData] = pydantic.Field(
        description="The list of (x,y,t) plots of the MHT problem."
    )
    misdetection_maximum_time_steps: pydantic.conint(ge=0) = pydantic.Field(
        default=0,
        description="The maximum number of time steps a target might be misdetected.",
    )
    penalty_energy: float = pydantic.Field(
        default=2,
        description="Penalty energy for invalid solutions. The value affects "
        "the converges rate. Small positive values are preferred",
    )
    three_local_coeff: float = pydantic.Field(
        default=0,
        description="Coefficient for the 3-local terms in the Hamiltonian. It is related to the angular acceleration.",
    )
    is_penalty: bool = pydantic.Field(
        default=True, description="Build Pubo using penalty terms"
    )
    max_velocity: float = pydantic.Field(
        default=0, description="Max allowed velocity for a segment"
    )

    def is_valid_cost(self, cost: float) -> bool:
        return True

    @pydantic.validator("plot_list")
    def round_plot_list_times_and_validate(cls, plot_list):
        MhtQaoaInput._check_all_ids_are_distinct(plot_list)
        MhtQaoaInput._round_to_tolerance_decimals(plot_list)

        time_stamps = sorted({plot.t for plot in plot_list})
        time_diff_set = {
            np.round(time_stamps[i] - time_stamps[i - 1], decimals=_TOLERANCE_DECIMALS)
            for i in range(1, len(time_stamps))
        }

        if len(time_diff_set) != 1:
            raise ValueError("The time difference between each time stamp is not equal")

        return plot_list

    @staticmethod
    def _round_to_tolerance_decimals(plot_list: List[PlotData]):
        for plot in plot_list:
            plot.t = np.round(plot.t, decimals=_TOLERANCE_DECIMALS)

    @staticmethod
    def _check_all_ids_are_distinct(plot_list: List[PlotData]):
        if not more_itertools.all_unique(plot.plot_id for plot in plot_list):
            raise ValueError("Plot IDs should be unique.")



@staticmethod
def _pprint_members(x):
    if x.isfinite():
        return "{" + str(sorted(x.ordered_data()))[1:-1] + "}"
    else:
        ans = " | ".join(str(_) for _ in x.ranges())
        if " | " in ans:
            return "(" + ans + ")"
        if ans:
            return ans
        else:
            return "[]"



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Functions for saving and loading Pyomo objects to json
"""
import datetime
import gzip
import json
import time

from classiq_interface.pyomo_extension import pyomo
from pyomo.core.base import _GeneralVarData
from pyomo.environ import Suffix

__format_version__ = 1

from classiq_interface.combinatorial_optimization.model_io_comon import (
    StoreSpec,
    may_have_subcomponents,
)


def _can_serialize(obj):
    try:
        json.dumps(obj)
        return True
    except TypeError:
        return False


class Counter(object):
    """
    This is a counter object, which is an easy way to pass an integer pointer
    around between methods.
    """

    def __init__(self):
        self.count = 0


def _may_have_expr(obj):
    """
    Args:
        obj: an object.
    Returns:
        True if the object has a expr attribute, otherwise
        False.
    """
    if hasattr(obj, "expr"):
        return True


def _may_have_args(obj):
    """
    Args:
        obj: an object.
    Returns:
        True if the object has a args attribute, otherwise
        False.
    """
    if hasattr(obj, "args"):
        return True


def _write_component(
    obj_dict: dict, obj, store_spec: StoreSpec, count=None, lookup=None, suffixes=None
):
    """
    Writes a component state to the save dictionary under a key given by the
    components name.
    Args:
        obj_dict: dictionary to to save the object into, will create a key that is the
            object name (not fully qualified)
        obj: object to save
        store_spec: a StoreSpec object indicating what object attributes to write
        count: count the number of Pyomo components written also used for ids
        lookup: is a lookup table for component ids from components
        suffixes: is a list of suffixes, that we are delaying writing
    Returns:
        None
    """
    if lookup is None:
        lookup = {}
    if suffixes is None:
        suffixes = []

    # Get the component name, doesn't need to be fully qualified or unique because
    # we are storing the state in a hierarchy structure
    if hasattr(obj, "getname"):
        obj_name = obj.getname(fully_qualified=False)
    else:
        obj_name = "noname"

    obj_dict[obj_name] = {"__type__": str(type(obj))}
    if store_spec.include_suffix:
        obj_dict[obj_name]["__id__"] = count.count
    lookup[id(obj)] = count.count
    if count is not None:
        count.count += 1

    attr_list, filter_function = store_spec.get_class_attr_list(obj)
    for attr in attr_list:
        if (
            attr in store_spec.get_functions
            and store_spec.get_functions[attr] is not None
        ):
            obj_dict[obj_name][attr] = store_spec.get_functions[attr](obj)
        else:
            obj_dict[obj_name][attr] = getattr(obj, attr, None)

    obj_dict[obj_name]["data"] = {}

    # if is a suffix, make a list and delay writing data until all components have an assigned id
    if isinstance(obj, Suffix):
        if store_spec.include_suffix:
            if store_spec.suffix_filter is None or obj_name in store_spec.suffix_filter:
                suffixes.append(
                    {
                        "obj_dict": obj_dict[obj_name]["data"],
                        "obj": obj,
                        "store_spec": store_spec,
                        "lookup": lookup,
                    }
                )

    else:
        _write_component_data(
            obj_dict=obj_dict[obj_name]["data"],
            obj=obj,
            store_spec=store_spec,
            count=count,
            lookup=lookup,
            suffixes=suffixes,
        )


def _write_component_data(
    obj_dict,
    obj,
    store_spec,
    count: Counter = None,
    lookup: dict = None,
    suffixes: list = None,
):
    """
    Iterate through the component data and write to the obj_dict dictionary. The keys
    for the data items are added to the dictionary. If the component has
    subcomponents they are written by a recursive call to _write_component under
    the __pyomo_components__ key.
    Args:
        obj_dict: dictionary to to save the object into, will create keys that are the
            data object indexes repn.
        obj: object to save
        store_spec: a StoreSpec object indicating what object attributes to write
        count: count the number of Pyomo components written also used for ids
        lookup: is a lookup table for component ids from components
        suffixes: is a list of suffixes, that we are delaying writing
    Returns:
        None
    """
    if lookup is None:
        lookup = {}
    if suffixes is None:
        suffixes = []

    if store_spec.include_suffix and isinstance(obj, Suffix):
        for key in obj:
            el = obj[key]
            if id(key) not in lookup:
                # didn't store these components so can't write suffix.
                continue
            if not _can_serialize(el):
                # Since I had the bright idea to expressions in suffixes
                # not everything in a suffix is serializable.
                continue
            obj_dict[lookup[id(key)]] = el  # Assume keys are Pyomo model components
        return

    try:
        item_keys = obj.keys()
    except AttributeError:
        item_keys = [None]

    attr_list = []

    for key in item_keys:
        if isinstance(obj, pyomo.core.base.set.Set):
            el = key
        else:
            try:
                el = obj[key]
            except TypeError:
                el = obj

        if not attr_list:  # assume all item are same type, use first to get attr_list
            attr_list, _ = store_spec.get_data_class_attr_list(el)

        repr_dict = {"__type__": str(type(el))}
        if store_spec.include_suffix:
            repr_dict["__id__"] = count.count
            lookup[id(el)] = count.count

        if count is not None:
            count.count += 1

        for attr in attr_list:  # store desired attributes
            if (
                attr in store_spec.get_functions
                and store_spec.get_functions[attr] is not None
            ):
                repr_dict[attr] = store_spec.get_functions[attr](el)
            else:
                repr_dict[attr] = getattr(el, attr)

        if may_have_subcomponents(el):  # block or block like component
            repr_dict["__pyomo_components__"] = {}
            for component in el.component_objects(descend_into=False):
                _write_component(
                    obj_dict=repr_dict["__pyomo_components__"],
                    obj=component,
                    store_spec=store_spec,
                    count=count,
                    lookup=lookup,
                    suffixes=suffixes,
                )

        if _may_have_expr(el):
            repr_dict["expr"] = {}
            _write_component(
                obj_dict=repr_dict["expr"],
                obj=el.expr,
                store_spec=store_spec,
                count=count,
                lookup=lookup,
                suffixes=suffixes,
            )

        elif _may_have_args(el):
            repr_dict["args"] = {}
            for idx_arg, arg in enumerate(el.args):
                repr_dict["args"][idx_arg] = {}
                _write_component(
                    obj_dict=repr_dict["args"][idx_arg],
                    obj=arg,
                    store_spec=store_spec,
                    count=count,
                    lookup=lookup,
                    suffixes=suffixes,
                )

        obj_dict[repr(key)] = repr_dict


def component_data_to_dict(obj, store_spec):
    """
    Component data to a dict.
    """
    attr_list, filter_function = store_spec.get_data_class_attr_list(obj)
    element_dict = {}  # if not writing suffixes don't need ids
    for attr in attr_list:  # store desired attributes
        element_dict[attr] = getattr(obj, attr)

    if may_have_subcomponents(obj):
        element_dict["__pyomo_components__"] = {}
        for component in obj.component_objects(descend_into=False):
            _write_component(
                obj_dict=element_dict["__pyomo_components__"],
                obj=component,
                store_spec=store_spec,
            )
    return element_dict


def to_json(
    obj,
    file_name=None,
    human_read=False,
    store_spec=None,
    metadata=None,
    gz=None,
    return_dict=False,
    return_json_string=False,
):
    """
    Save the state of a model to a Python dictionary, and optionally dump it
    to a json file.  To load a model state, a model with the same structure must
    exist.  The model itself cannot be recreated from this.
    Args:
        obj: The Pyomo component object to save.  Usually a Pyomo model, but could
            also be a subcomponent of a model (usually a sub-block).
        file_name: json file name to save model state, if None only create
            python dict
        gz: If file_name is given and gv is True gzip the json file. The default is
            True if the file name ends with '.gz' otherwise False.
        human_read: if True, add indents and spacing to make the json file more
            readable, if false cut out whitespace and make as compact as
            possible
        metadata: A dictionary of additional metadata to add.
        store_spec: is What To Save, this is a StoreSpec object that specifies what
            object types and attributes to save.  If None, the default is used
            which saves the state of the complete model state.
        metadata: additional metadata to save beyond the standard format_version,
            date, and time.
        return_dict: default is False if true returns a dictionary representation
        return_json_string: default is False returns a json string
    Returns:
        If return_dict is True returns a dictionary serialization of the Pyomo
        component.  If return_dict is False and return_json_string is True
        returns a json string dump of the dict.  If file_name is given the dictionary
        is also written to a json file.  If gz is True and file_name is given, writes
        a gzipped json file.
    """
    if gz is None:
        if isinstance(file_name, str):
            gz = file_name.endswith(".gz")
        else:
            gz = False
    if metadata is None:
        metadata = {}

    suffixes = []
    lookup = {}
    count = Counter()
    start_time = time.time()
    if store_spec is None:
        store_spec = StoreSpec()

    now = datetime.datetime.now()
    obj_dict = {
        "__metadata__": {
            "format_version": __format_version__,
            "date": datetime.date.isoformat(now.date()),
            "time": datetime.time.isoformat(now.time()),
            "other": metadata,
        }
    }

    _write_component(obj_dict, obj, store_spec, count, suffixes=suffixes, lookup=lookup)
    for s in suffixes:
        _write_component_data(**s)

    obj_dict["__metadata__"]["__performance__"] = {}
    performance_dict = obj_dict["__metadata__"]["__performance__"]
    performance_dict["n_components"] = count.count
    dict_time = time.time()
    performance_dict["time_to_make_dict"] = dict_time - start_time
    dump_kw = {"indent": 2} if human_read else {"separators": (",", ":")}
    if file_name is not None:
        if gz:
            with gzip.open(file_name, "w") as f:
                f.write(json.dumps(obj_dict, **dump_kw).encode("utf-8"))
        else:
            with open(file_name, "w") as f:
                json.dump(obj_dict, f, **dump_kw)
    file_time = time.time()
    performance_dict["time_to_write_file"] = file_time - dict_time

    if return_dict:
        return obj_dict
    elif return_json_string:
        return json.dumps(obj_dict, **dump_kw)
    else:
        return None



import enum
from typing import Optional, List, Union

import pydantic

from classiq_interface.generator.generation_metadata import GenerationMetadata
from classiq_interface.generator.noise_properties import NoiseProperties
from classiq_interface.backend.backend_preferences import (
    AwsBackendPreferences,
    IBMBackendPreferences,
    AzureBackendPreferences,
)


class VerbosityLevel(str, enum.Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"


ANSATZ_EXTENSION = ".qasm"
COST_DATA_EXTENSION = ".json"


class CostType(str, enum.Enum):
    MIN = "MIN"
    AVERAGE = "AVERAGE"
    CVAR = "CVAR"


class VQEPreferences(pydantic.BaseModel):
    qaoa_reps: pydantic.PositiveInt = pydantic.Field(
        default=1, description="Number of layers in qaoa ansatz."
    )
    num_shots: pydantic.PositiveInt = pydantic.Field(
        default=100, description="Number of repetitions of the quantum ansatz."
    )
    cost_type: CostType = pydantic.Field(
        default=CostType.CVAR,
        description="Summarizing method of the measured bit strings",
    )
    external_cost_server: Optional[pydantic.AnyHttpUrl] = pydantic.Field(
        default=None,
        description="URL for external cost server. Format: http://<host>:<port>/score "
        "The server must accept POST requests with field 'trk_list' and return"
        "a response with field 'score'.",
    )
    alpha_cvar: pydantic.confloat(gt=0, le=1) = pydantic.Field(
        default=None, description="Parameter for the CVAR summarizing method"
    )
    max_iteration: pydantic.PositiveInt = pydantic.Field(
        default=100, description="Maximal number of optimizer iterations"
    )
    tolerance: pydantic.PositiveFloat = pydantic.Field(
        default=None, description="Final accuracy in the optimization"
    )
    verbosity_level: VerbosityLevel = pydantic.Field(
        default=VerbosityLevel.DEBUG,
        description="Level of elaboration of the presented output. "
        "INFO stands for critical data only. "
        "DEBUG stands for extended data about the performance.",
    )
    num_presented_solutions: Optional[pydantic.PositiveInt] = pydantic.Field(
        default=3,
        description="Number of presented "
        "solutions. Set to null to "
        "fetch all solutions.",
    )
    use_gpu: bool = pydantic.Field(
        default=True, description="Try simulating on GPU if available."
    )
    invalid_space_cost_threshold: Optional[pydantic.PositiveFloat] = pydantic.Field(
        default=None,
        description="Solutions with cost greater or equal to that value, are considered invalid.",
    )
    is_penalty: bool = pydantic.Field(
        default=True,
        description="Indicates whether to use QAOA with penalty terms or constrained QAOA.",
    )
    penalty_energy: float = pydantic.Field(
        default=2,
        description="Penalty energy for invalid solutions. The value affects "
        "the converges rate. Small positive values are preferred",
    )

    @pydantic.validator("alpha_cvar", pre=True, always=True)
    def check_alpha_cvar(cls, alpha_cvar, values):
        if alpha_cvar is not None and values["cost_type"] != CostType.CVAR:
            raise ValueError("Use CVAR params only when relevant.")

        if alpha_cvar is None and values["cost_type"] == CostType.CVAR:
            alpha_cvar = 0.04

        return alpha_cvar


class VQEProblem(pydantic.BaseModel):
    # TODO: add random seed
    vqe_preferences: VQEPreferences = pydantic.Field(
        default_factory=VQEPreferences, description="preferences for the VQE execution"
    )
    ansatz: str = pydantic.Field(
        default=None, description="Qasm containing the ansatz circuit"
    )
    cost_data: GenerationMetadata = pydantic.Field(
        default=None, description="Data returned from the generation procedure."
    )
    is_generate_strings_uniformly: bool = pydantic.Field(
        default=True,
        description="Change the parameter sampling distribution in order to get uniform "
        "distribution on the valid bit strings.",
    )
    warm_start: Optional[List[float]] = pydantic.Field(
        default=None,
        description="Supply a vector of parameter assignments "
        "as an initial guess for the VQE search",
    )
    noise_properties: Optional[NoiseProperties] = pydantic.Field(
        default=None, description="Properties of the noise in the circuit"
    )
    use_mps_simulator: bool = pydantic.Field(
        default=False,
        description="Whether to use a matrix_product_state instead"
        "of statevector simulator. Use to improve"
        "performance of lightly entangled circuits.",
    )
    backend_preferences: Union[
        AzureBackendPreferences, IBMBackendPreferences, AwsBackendPreferences
    ] = pydantic.Field(
        default_factory=lambda: IBMBackendPreferences(
            backend_service_provider="IBMQ", backend_name="aer_simulator"
        ),
        description="Preferences for the requested backend to run the quantum circuit.",
    )

    @pydantic.validator("warm_start")
    def check_num_of_params_assign(cls, warm_start, values):
        # TODO: deprecate completely
        # if warm_start is None:
        #     return warm_start
        #
        # ansatz = values.get("ansatz")
        # circuit_param = qiskit.QuantumCircuit.from_qasm_str(ansatz).parameters
        #
        # if len(warm_start) != len(circuit_param):
        #     raise ValueError(
        #         "Number of parameters assigned does not match number of parameters in circuit"
        #     )

        return warm_start



from typing import List, Literal, Optional

import numpy as np
import pydantic

from classiq_interface.generator import adjacency
from classiq_interface.hybrid.problem_input import ProblemInput, OptimizationProblemName


class MVCInput(ProblemInput):
    name: Literal[OptimizationProblemName.MVC] = pydantic.Field(
        default=OptimizationProblemName.MVC, description="Name of optimization problem."
    )
    n: pydantic.PositiveInt = pydantic.Field(
        default=..., description="Number of vertices in the graph."
    )
    k: pydantic.PositiveInt = pydantic.Field(
        default=..., description="Size of cover vertices subset."
    )
    adjacency_matrix: Optional[List[List[bool]]] = pydantic.Field(
        default=None, description="Graph adjacency matrix."
    )
    reps: pydantic.conint(ge=1) = pydantic.Field(
        default=3, description="Number of QAOA layers."
    )

    @pydantic.validator("adjacency_matrix", pre=True, always=True)
    def check_adjacency_matrix(cls, matrix, values):
        if matrix is None:
            matrix = adjacency.get_rand_adjacency_matrix(num_vertices=values["n"])
        array = np.array(matrix)
        if not array.shape[0] == array.shape[1]:
            raise ValueError("distance matrix must be square")
        if not np.allclose(array, array.T):
            raise ValueError("Adjacency matrix is not symmetric")
        if np.diag(array).any():
            raise ValueError("Diagonal of adjacency matrix is not zero")
        return matrix

    def is_valid_cost(self, cost: float) -> bool:
        adjacency_matrix = np.array(self.adjacency_matrix)
        num_edges = np.sum(adjacency_matrix) / 2
        _MIN_COST = 0

        return _MIN_COST <= cost <= num_edges



from enum import Enum
from typing import Type, Optional

import pydantic

from classiq_interface.generator.function_params import FunctionParams


class McxInputs(Enum):
    CTRL_IN = "CTRL_IN"
    TARGET_QUBIT_IN = "TARGET_QUBIT"


class McxOutputs(Enum):
    CTRL_OUT = "CTRL_OUT"
    TARGET_QUBIT_OUT = "TARGET_QUBIT_OUT"


class Mcx(FunctionParams):
    """
    multi-controlled x-gate
    """

    num_ctrl_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of control qubits."
    )
    ctrl_state: Optional[str] = pydantic.Field(
        default=None, description="string of the control state"
    )

    @pydantic.validator("ctrl_state", always=True)
    def validate_ctrl_state(cls, ctrl_state, values):
        num_ctrl_qubits = values.get("num_ctrl_qubits")
        if ctrl_state is None:
            return "1" * num_ctrl_qubits

        if len(ctrl_state) != num_ctrl_qubits:
            raise ValueError(
                "control state length should be equal to the number of control qubits"
            )
        return ctrl_state

    _input_names: Type[Enum] = pydantic.PrivateAttr(default=McxInputs)
    _output_names: Type[Enum] = pydantic.PrivateAttr(default=McxOutputs)



from enum import Enum
from typing import Type
from abc import ABC

import pydantic

from classiq_interface.generator.function_params import FunctionParams


class EntanglerInputs(Enum):
    IN = "IN"


class EntanglerOutputs(Enum):
    OUT = "OUT"


class Entangler(FunctionParams, ABC):
    """
    A Father class for all entangler classes
    """

    qubit_count: pydantic.PositiveInt = pydantic.Field(
        description="The number of qubits for the entangler."
    )

    schmidt_rank: pydantic.conint(ge=0) = pydantic.Field(
        default=0, description="The required schmidt rank (log of schmidt number)."
    )
    is_input_zero: bool = pydantic.Field(
        default=True,
        description="Boolean determining whether the entangler must "
        "act on state |0...0>.",
    )

    def create_io_enums(self):
        if not self.is_input_zero:
            self._input_names = EntanglerInputs
            self._output_names = EntanglerOutputs


class TwoDimensionalEntangler(Entangler):
    """
    Creates a two dimensional cluster state with the specified number of qubits and schmidt rank
    (log of schmidt number). When the desired schmidt rank is too high, a rectangular grid with schmidt rank
    floor(sqrt(qubit_count))-1 is generated.
    """

    pass


class HypercubeEntangler(Entangler):
    """
    Creates a cluster/graph state in the form of a hypercube with the specified number of qubits. The hypercube is
    constructed by building cubes of growing dimension therefore if the number of qubits is not a a power of 2 (n=2^k)
    the last cube will not be completed. for example if n = 11 = 2^3 + 3 a three dimensional cube is constructed
    connected to additional 3 qubits in the natural order
    (that is, these qubits will be: 1000, 1001, 1010)
    """

    pass


class GridEntangler(Entangler):
    """
    creates a graph state in the form of multi-dimensional grid according to the specified number of qubits and Schmidt
    rank. If possible the grid will include the exact Schmidt rank if not a smaller grid with a lower schmidt rank is
    constructed - as close as possible to the specified parameters. if the specified Schmidt rank is too high a 'long'
    grid with the maximal possible Schmidt rank width is constructed (that still obeys the condition that the largest
    dimension minus 1 is larger then the sum of the (d_i - 1) -- d_i including all other dimensions)
    """

    grid_randomization: bool = pydantic.Field(
        default=True,
        description="Boolean determining whether the grid structure is randomly selected out of all grids which provide"
        "the same Schmidt rank width. If False the grid with maximal number of dimensions is selected",
    )



from typing import List

import pydantic

from classiq_interface.generator.circuit_outline import Qubit


class SatConstraint(pydantic.BaseModel):
    variables: pydantic.conlist(item_type=Qubit, min_items=1)
    truth_table: List[bool]

    @pydantic.validator("truth_table")
    def truth_table_validator(cls, truth_table):
        if not any(truth_table):
            raise ValueError("truth table must contain at least 1 true value")

        return truth_table

    @pydantic.root_validator(skip_on_failure=True)
    def combinations_length_validator(cls, values):
        variables, truth_table = values.get("variables"), values.get("truth_table")

        assert (
            variables is not None and truth_table is not None
        ), "These can not be none, as skip_on_failure is True."

        possible_combinations_count = 2 ** len(variables)
        if len(truth_table) != possible_combinations_count:
            raise ValueError(
                f"length of truth_table truth table must be {possible_combinations_count}"
            )

        return values


class SatProblem(pydantic.BaseModel):
    variables_count: pydantic.PositiveInt
    constraints: List[SatConstraint]



import pydantic
from typing import TypeVar

Qubit = TypeVar("Qubit", bound=pydantic.conint(ge=0))
Cycle = TypeVar("Cycle", bound=pydantic.conint(ge=0))
Clbit = TypeVar("Clbit", bound=pydantic.conint(ge=0))



from collections import defaultdict
from typing import Dict, Optional, List, Union

import classiq_interface.generator.validations.flow_graph as flow_graph
import networkx as nx
import pydantic
from classiq_interface.generator.barriers import (
    SingleRegisterBarrier,
    SingleQubitBarrier,
)
from classiq_interface.generator.classical_if import Cif
from classiq_interface.generator.sat_problem import (
    SatProblem,
)
from classiq_interface.generator.custom_gates import CustomGate, ValidGateName
from classiq_interface.generator.entanglers import (
    CXSlideEntangler,
    PhaseCXSlideEntangler,
)
from classiq_interface.generator.measurements import Measurement
from classiq_interface.generator.phase_estimation import PhaseEstimation
from classiq_interface.generator.preferences.optimization import Optimization
from classiq_interface.generator.preferences.randomness import create_random_seed
from classiq_interface.generator.qiskit_quantum_gates import QiskitBuiltinQuantumGates
from classiq_interface.generator.range_types import NonNegativeIntRange
from classiq_interface.generator.segments import FunctionCall
from classiq_interface.generator.slides import PhaseSlide
from classiq_interface.generator.teleport import Teleport
from classiq_interface.generator.transpilation import TranspilationPreferences
from classiq_interface.generator.result import QuantumFormat

DEFAULT_MINIMAL_DEPTH = 1
CYCLE_ERROR_MSG = "Inputs and outputs cannot form a cycle"


# TODO define a type that can be used in variable declaration that is consistent with usage
def normalize_dict_key_to_str(d):
    return {k.name: v for k, v in d.items()}


class QuantumCircuitConstraints(pydantic.BaseModel):
    """
    Input constraints for the generated quantum circuit.
    """

    # TODO: Consider moving timeout outside of constraints, and supply it (optionally) separate of the constraints.
    # TODO: Remove hard coded timeout when issue,https://github.com/MiniZinc/minizinc-python/pull/8 is resolved
    timeout_seconds: Optional[pydantic.PositiveInt] = pydantic.Field(
        default=300, description="Generation timeout in seconds"
    )
    qubit_count: pydantic.PositiveInt = pydantic.Field(
        default=...,
        description="Number of qubits in generated quantum circuit",
    )
    min_depth: pydantic.PositiveInt = DEFAULT_MINIMAL_DEPTH
    max_depth: pydantic.PositiveInt
    random_seed: int = pydantic.Field(
        default_factory=create_random_seed,
        description="The random seed used for the generation",
    )
    gate_count_constraints: Dict[
        QiskitBuiltinQuantumGates, NonNegativeIntRange
    ] = pydantic.Field(default_factory=lambda: defaultdict(NonNegativeIntRange))
    custom_gate: Optional[Dict[ValidGateName, CustomGate]] = None
    cx_slide_entangler: Optional[CXSlideEntangler] = None
    phase_cx_slide_entangler: Optional[PhaseCXSlideEntangler] = None
    phase_slide: Optional[PhaseSlide] = None
    logic_flow: List[FunctionCall] = pydantic.Field(
        default_factory=list,
        description="List of function calls to be applied in the circuit",
    )
    combine_segments: bool = pydantic.Field(
        default=False,
        description="If set to true, all segments will be combined into one"
        "segment. The way they are connected and the qubit assignment "
        "will be set deterministically, without using the CSP model.",
    )
    allowed_gates: List[QiskitBuiltinQuantumGates] = pydantic.Field(
        default_factory=list
    )
    phase_estimation: Optional[PhaseEstimation] = None
    teleport: Optional[Teleport] = None
    optimization: Optional[Optimization] = pydantic.Field(default_factory=Optimization)
    single_measurements: List[Measurement] = pydantic.Field(default_factory=list)
    register_barriers: List[SingleRegisterBarrier] = pydantic.Field(
        default_factory=list
    )
    qubit_barriers: List[SingleQubitBarrier] = pydantic.Field(default_factory=list)
    single_cifs: List[Cif] = pydantic.Field(default_factory=list)
    sat_problem: Optional[SatProblem] = None
    two_qubits_gate_count: NonNegativeIntRange = pydantic.Field(
        default_factory=NonNegativeIntRange
    )
    transpilation: Optional[TranspilationPreferences] = pydantic.Field(
        default_factory=TranspilationPreferences
    )
    segment_qubit_count: bool = pydantic.Field(
        default=False,
        description="If set to True, the number of qubits will"
        " be determined by the width of the single"
        " segment in the circuit",
    )
    output_format: Union[
        QuantumFormat,
        pydantic.conlist(
            item_type=QuantumFormat,
            min_items=1,
            max_items=len(QuantumFormat),
        ),
    ] = pydantic.Field(
        default=[QuantumFormat.QASM],
        description="The quantum circuit output format, qasm by default. Q# (qs) is "
        "also supported. It can also be a list of unique quantum formats. In the case "
        "of multiple formats, all formats will be saved and the first one will be "
        "presented.",
    )
    use_synthesis_engine: bool = False

    _gate_count_constraints = pydantic.validator(
        "gate_count_constraints", allow_reuse=True
    )(normalize_dict_key_to_str)

    class Config:
        extra = "forbid"

    @pydantic.validator("max_depth")
    def validate_max_depth(cls, max_depth, values):
        min_depth = values.get("min_depth")
        if min_depth and max_depth < min_depth:
            raise ValueError("max_depth must be greater or equal to min_depth")

        return max_depth

    @pydantic.validator("allowed_gates")
    def normalize_list_key_to_str(cls, use_gate):
        return [v.name for v in use_gate]

    @pydantic.validator("logic_flow")
    def validate_segments(cls, segment_list):
        if not segment_list:
            return segment_list

        graph = flow_graph.create_flow_graph(segment_list)

        if not nx.algorithms.is_directed_acyclic_graph(graph):
            cycles = list(nx.algorithms.simple_cycles(graph))
            raise ValueError(CYCLE_ERROR_MSG + ". Cycles are: " + str(cycles))

        return segment_list

    @pydantic.validator("segment_qubit_count")
    def validate_segment_qubit_count(cls, segment_qubit_count, values):
        if not segment_qubit_count:
            return segment_qubit_count

        segments = values.get("logic_flow")
        if segments is None:
            return segment_qubit_count
        if len(segments) == 1:
            return segment_qubit_count

        combine_segments = values.get("combine_segments")
        if combine_segments:
            return segment_qubit_count

        raise ValueError(
            "Segment qubit count can only be true if there is only one segment"
        )

    @pydantic.validator("output_format")
    def validate_output_format(cls, output_format):
        if isinstance(output_format, QuantumFormat):
            return [output_format]
        else:
            if len(output_format) == len(set(output_format)):
                return output_format
            else:
                raise ValueError(
                    f"{output_format=}\n"
                    "has at least one format that appears twice or more"
                )



import enum


class QiskitBuiltinQuantumGates(str, enum.Enum):
    # TODO: Not using 'I' due to https://www.flake8rules.com/rules/E741.html
    I = "I"
    X = "X"
    Y = "Y"
    Z = "Z"
    T = "T"
    S = "S"
    H = "H"
    P = "P"
    CX = "CX"
    CY = "CY"
    CZ = "CZ"
    CCX = "CCX"
    RY = "RY"
    CRY = "CRY"



import abc
from enum import Enum
from typing import Tuple, Union

import pydantic

from classiq_interface.generator import convert_phase, probability
from classiq_interface.generator.circuit_outline import Qubit


class Slide(pydantic.BaseModel, abc.ABC):
    """
    A sequence of same-type-constructs, on a sequence of qubits.
    """

    # TODO: Verify values are not greater than max index.
    slide_targets: Tuple[Qubit, ...] = pydantic.Field(
        default=..., description="The sequence of qubits to assign the construct on."
    )

    @pydantic.validator("slide_targets")
    def slide_targets_uniqueness(cls, value):
        if len(value) != len(set(value)):
            raise ValueError("must not contain repetitions")

        return value


class PhaseDistributionType(str, Enum):
    UNIFORM = "uniform"
    RANDOM = "random"


class PhaseDistribution(pydantic.BaseModel):
    distribution_type: PhaseDistributionType
    total_phase: pydantic.confloat(ge=0)
    # The phase here is converted to degrees within phase slide.


class PhaseSlide(Slide):
    slide_phases: Union[PhaseDistribution, Tuple[float, ...], Tuple[int, ...]]

    @pydantic.validator("slide_phases", always=True)
    def phases_validator(cls, slide_phases, values):
        slide_targets = values.get("slide_targets")
        if slide_targets is None:
            # In that case, don't do any validation.
            return slide_phases

        if isinstance(slide_phases, PhaseDistribution):
            slide_phases = cls.distribute_phase(
                phase_distribution=slide_phases, entries_count=len(slide_targets)
            )

        if len(slide_phases) != len(slide_targets):
            raise ValueError("phases and targets length mismatch")

        slide_phases_rad: Tuple[int, ...] = tuple(
            convert_phase.rad_to_deg(i) for i in slide_phases
        )

        return slide_phases_rad

    @staticmethod
    def distribute_phase(
        phase_distribution: PhaseDistribution, entries_count: int
    ) -> Tuple[float, ...]:
        if phase_distribution.distribution_type == PhaseDistributionType.UNIFORM:
            return (
                convert_phase.to_canonical_phase(
                    phase_distribution.total_phase / entries_count
                ),
            ) * entries_count

        if phase_distribution.distribution_type == PhaseDistributionType.RANDOM:
            random_probability = probability.random_probability(
                options_count=entries_count
            )
            return tuple(
                convert_phase.to_canonical_phase(phase_distribution.total_phase * prob)
                for prob in random_probability
            )

        assert (
            False
        ), f"non supported distribution type {phase_distribution.distribution_type}"



from typing import Set, Type

from classiq_interface.generator.arithmetic import (
    BitwiseAnd,
    BitwiseOr,
    BitwiseXor,
    Adder,
    Arithmetic,
    Equal,
    ArithmeticOracle,
    Negation,
)
from classiq_interface.generator.finance import Finance
from classiq_interface.generator.function_params import FunctionParams
from classiq_interface.generator.grover_operator import GroverOperator
from classiq_interface.generator.hadamard_amp_load import HadamardAmpLoad
from classiq_interface.generator.integer_comparator import IntegerComparator
from classiq_interface.generator.qaoa_ansatz import QaoaAnsatz
from classiq_interface.generator.qft import QFT
from classiq_interface.generator.sparse_amp_load import SparseAmpLoad
from classiq_interface.generator.state_preparation import StatePreparation
from classiq_interface.generator.state_propagator import StatePropagator
from classiq_interface.generator.entangler_params import TwoDimensionalEntangler
from classiq_interface.generator.vqe_ansatz import VQEAnsatz
from classiq_interface.generator.amplitude_estimation import AmplitudeEstimation
from classiq_interface.generator.entangler_params import HypercubeEntangler
from classiq_interface.generator.entangler_params import GridEntangler
from classiq_interface.generator.mcx import Mcx

_function_param_list = {
    StatePreparation,
    VQEAnsatz,
    QaoaAnsatz,
    StatePropagator,
    QFT,
    BitwiseAnd,
    BitwiseOr,
    BitwiseXor,
    Adder,
    Arithmetic,
    ArithmeticOracle,
    Equal,
    Negation,
    TwoDimensionalEntangler,
    IntegerComparator,
    Finance,
    HypercubeEntangler,
    AmplitudeEstimation,
    SparseAmpLoad,
    GridEntangler,
    HadamardAmpLoad,
    GroverOperator,
    Mcx,
}


def get_function_param_list() -> Set[Type[FunctionParams]]:
    return _function_param_list



from typing import Dict, Union
import pydantic

from classiq_interface.backend.backend_preferences import (
    IBMBackendPreferences,
    AwsBackendPreferences,
    AzureBackendPreferences,
)
from classiq_interface.hybrid.vqe_problem import VQEPreferences


class OptimizationProblem(pydantic.BaseModel):
    vqe_preferences: VQEPreferences = pydantic.Field(
        default_factory=VQEPreferences, description="preferences for the VQE execution"
    )
    serialized_model: Dict = None
    backend_preferences: Union[
        AzureBackendPreferences, IBMBackendPreferences, AwsBackendPreferences
    ] = pydantic.Field(
        default_factory=lambda: IBMBackendPreferences(
            backend_service_provider="IBMQ", backend_name="aer_simulator"
        ),
        description="Preferences for the requested backend to run the quantum circuit.",
    )



import pydantic
from typing import Optional, Literal


class NoiseProperties(pydantic.BaseModel):

    # Must be synced with hybrid.hardware_noise_models.hardware_noise_dict
    _HARDWARE_NAMES = (
        "ibmq_16_melbourne",
        "ibmq_armonk",
        "ibmq_athens",
        "ibmq_belem",
        "ibmq_lima",
        "ibmq_quito",
        "ibmq_santiago",
        "ibmqx2",
        "ibmq_manila",
    )

    measurement_bit_flip_probability: Optional[
        pydantic.confloat(ge=0, le=1)
    ] = pydantic.Field(
        default=None,
        description="Probability of measuring the wrong value for each qubit.",
    )
    hardware_noise_name: Optional[Literal[_HARDWARE_NAMES]] = pydantic.Field(
        default=None, description="Name of hardware to simulate its noise"
    )

    @pydantic.validator("hardware_noise_name")
    def validate_hardware_noise_simulator_alone(cls, hardware_noise_name, values):
        if not hardware_noise_name:
            return

        # Hardware noise model cannot work with more noise
        if any(v is not None for v in values.values()):
            raise ValueError(
                "Hardware noise model cannot be changed with additional noise."
            )

        return hardware_noise_name



from enum import Enum
from typing import Type

import pydantic

from classiq_interface.generator import function_params


class QftInputs(Enum):
    IN = "IN"


QftOutputs = function_params.DefaultOutputNames


class QFT(function_params.FunctionParams):
    """
    Creates a quantum Fourier transform on a specified number of qubits.
    Use the inverse flag to create the inverse QFT circuit.
    """

    num_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of qubits on which the QFT acts."
    )
    approximation_degree: pydantic.conint(ge=0) = pydantic.Field(
        default=0,
        description="The degree of approximation (0 for no approximation). The smallest "
        "'approximation_degree' rotation angles are dropped from the QFT.",
    )
    do_swaps: bool = pydantic.Field(
        default=True, description="Whether to include the final swaps in the QFT."
    )
    inverse: bool = pydantic.Field(
        default=False,
        description="If True, the inverse Fourier transform is constructed.",
    )

    _input_names: Type[Enum] = pydantic.PrivateAttr(default=QftInputs)
    _output_names: Type[Enum] = pydantic.PrivateAttr(default=QftOutputs)



from enum import Enum
from typing import Tuple, Optional, Dict, Union

import pydantic
from classiq_interface.generator.function_params import (
    FunctionParams,
    DefaultOutputNames,
    DefaultInputNames,
)
from classiq_interface.generator.preferences.optimization import (
    StatePrepOptimizationMethod,
)
from classiq_interface.generator.range_types import NonNegativeFloatRange


class Metrics(str, Enum):
    KL = "KL"
    L2 = "L2"
    L1 = "L1"
    MAX_PROBABILITY = "MAX_PROBABILITY"


def is_power_of_two(pmf):
    n = len(pmf)
    is_power_of_two = (n != 0) and (n & (n - 1) == 0)
    if not is_power_of_two:
        raise ValueError("Probabilities length must be power of 2")
    return pmf


class PMF(pydantic.BaseModel):
    pmf: Tuple[pydantic.confloat(ge=0, le=1), ...]

    @pydantic.validator("pmf")
    def is_sum_to_one(cls, pmf):
        # n = len(pmf)
        # is_power_of_two = (n != 0) and (n & (n - 1) == 0)
        # if not is_power_of_two:
        #     raise ValueError("Probabilities length must be power of 2")
        if round(sum(pmf), 8) != 1:
            raise ValueError("Probabilities do not sum to 1")
        return pmf

    _is_pmf_valid = pydantic.validator("pmf", allow_reuse=True)(is_power_of_two)


class GaussianMoments(pydantic.BaseModel):
    mu: float
    sigma: pydantic.confloat(gt=0)


class GaussianMixture(pydantic.BaseModel):
    gaussian_moment_list: Tuple[GaussianMoments, ...]


StatePreparationOutputs = DefaultOutputNames
StatePreparationInputs = DefaultInputNames


class StatePreparation(FunctionParams):
    probabilities: Union[PMF, GaussianMixture]
    depth_range: Optional[NonNegativeFloatRange] = NonNegativeFloatRange(
        lower_bound=0, upper_bound=1e100
    )
    cnot_count_range: Optional[NonNegativeFloatRange] = NonNegativeFloatRange(
        lower_bound=0, upper_bound=1e100
    )
    error_metric: Optional[Dict[Metrics, NonNegativeFloatRange]] = pydantic.Field(
        default_factory=lambda: {
            Metrics.KL: NonNegativeFloatRange(lower_bound=0, upper_bound=1e100)
        }
    )
    optimization_method: Optional[
        StatePrepOptimizationMethod
    ] = StatePrepOptimizationMethod.KL
    num_qubits: Optional[int] = None
    is_uniform_start: bool = True

    @pydantic.validator("is_uniform_start")
    def is_uniform_start_validator(cls, is_uniform_start, values):
        if not is_uniform_start:
            if Metrics.KL in values["error_metric"]:
                raise ValueError("Error Metric for non-uniform start can be L1 ans L2")
            if StatePrepOptimizationMethod.KL in values["optimization_method"]:
                raise ValueError(
                    "Optimization method for non-uniform start can be L1 ans L2"
                )
        return is_uniform_start

    @pydantic.root_validator()
    def validate_num_qubits(cls, values):
        num_qubits = values["num_qubits"]
        probabilities = values["probabilities"]
        if isinstance(probabilities, GaussianMixture) and num_qubits is None:
            raise ValueError("num_qubits must be set when using gaussian mixture")

        return values

    class Config:
        extra = "forbid"



import networkx as nx
import pyomo.environ as pyo


def mvc_pyomo_model(graph: nx.Graph, k: int) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()
    model.Nodes = pyo.Set(initialize=list(graph.nodes))
    model.Arcs = pyo.Set(initialize=list(graph.edges))
    model.x = pyo.Var(model.Nodes, domain=pyo.Binary)
    model.amount_constraint = pyo.Constraint(expr=sum(model.x[i] for i in model.x) == k)

    def obj_expression(model):
        # number of edges not covered
        return sum((1 - model.x[i]) * (1 - model.x[j]) for i, j in model.Arcs)

    model.cost = pyo.Objective(rule=obj_expression, sense=pyo.minimize)

    return model



ROUTE_PREFIX = "/api/v1"

ANALYZER_PREFIX = ROUTE_PREFIX + "/analyzer"
HYBRID_PREFIX = ROUTE_PREFIX + "/hybrid"
EXECUTE_PREFIX = ROUTE_PREFIX + "/execute"
COMBINATORIAL_OPTIMIZATION_PREFIX = ROUTE_PREFIX + "/combinatorial_optimization"

TASKS_SUFFIX = "/tasks"

TASKS_GENERATE_SUFFIX = "/tasks/generate"
TASKS_SOLVE_SUFFIX = "/tasks/solve"

ANALYZER_FULL_PATH = ANALYZER_PREFIX + TASKS_SUFFIX
HYBRID_FULL_PATH = HYBRID_PREFIX + TASKS_SUFFIX
EXECUTE_FULL_PATH = EXECUTE_PREFIX + TASKS_SUFFIX
COMBINATORIAL_OPTIMIZATION_GENERATE_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_GENERATE_SUFFIX
)
COMBINATORIAL_OPTIMIZATION_SOLVE_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_SOLVE_SUFFIX
)



from typing import List

import numpy as np


# Taken and modified from quantum_tsp_tutorials repository of mstechly in github.
# https://github.com/mstechly/quantum_tsp_tutorials


def _distance_between_points(point_1: np.ndarray, point_2: np.ndarray) -> float:
    return np.linalg.norm(point_1 - point_2)


def get_distance_matrix(cities: np.ndarray) -> List[List[float]]:
    number_of_cities = len(cities)
    matrix = np.zeros((number_of_cities, number_of_cities))
    for i in range(number_of_cities):
        for j in range(i, number_of_cities):
            matrix[i][j] = _distance_between_points(cities[i], cities[j])
            matrix[j][i] = matrix[i][j]
    return matrix.tolist()


def get_rand_euclidean_distance_matrix(num_points: int) -> List[List[float]]:
    _POINTS_ARRAY_DIM = 2
    points = np.random.rand(num_points, _POINTS_ARRAY_DIM)
    distance_matrix = get_distance_matrix(points)
    return distance_matrix


def get_rand_distance_matrix(num_points: int) -> List[List[float]]:
    distance_matrix = np.random.rand(num_points, num_points)
    distance_matrix = (distance_matrix + distance_matrix.T) / 2
    distance_matrix -= np.diag(np.diag(distance_matrix))
    return distance_matrix.tolist()



import pydantic

from classiq_interface.generator import problem_name_mapping
from classiq_interface.hybrid.encoding_type import EncodingType
from classiq_interface.hybrid.problem_input import ProblemInput


class ProblemProperties(pydantic.BaseModel):
    problem: ProblemInput = pydantic.Field(
        default=..., description="Problem input data"
    )
    encoding_type: EncodingType = pydantic.Field(
        default=None, description="Various encoding types to the combinatorial problem."
    )

    @pydantic.validator("problem", pre=True)
    def check_problem(cls, problem):
        if isinstance(problem, dict) and problem.get("name") is None:
            return problem

        return problem_name_mapping.parse_problem_input(problem)

    @pydantic.validator("encoding_type", always=True)
    def check_encoding_type(cls, encoding_type, values):
        if encoding_type:
            return encoding_type
        problem = values.get("problem")
        if not problem:
            return encoding_type
        encoding_type = problem_name_mapping.ENCODING_TYPE_DEFAULT[type(problem)]
        return encoding_type



import enum
from typing import Dict, Type

import pydantic

from classiq_interface.generator.ansatz_library import (
    TwoLocalArgs,
    SeparateU3Args,
    HypercubeArgs,
    EntanglingLayersArgs,
    RandomArgs,
    RandomTwoQubitGatesArgs,
)


class CustomAnsatzType(str, enum.Enum):
    TwoLocal = "TwoLocal"
    SeparateU3 = "SeparateU3"
    Hypercube = "Hypercube"
    EntanglingLayers = "EntanglingLayers"
    Random = "Random"
    RandomTwoQubitGates = "RandomTwoQubitGates"


CUSTOM_ANSATZ_ARGS_MAPPING: Dict[CustomAnsatzType, Type[pydantic.BaseModel]] = {
    CustomAnsatzType.TwoLocal: TwoLocalArgs,
    CustomAnsatzType.SeparateU3: SeparateU3Args,
    CustomAnsatzType.Hypercube: HypercubeArgs,
    CustomAnsatzType.EntanglingLayers: EntanglingLayersArgs,
    CustomAnsatzType.Random: RandomArgs,
    CustomAnsatzType.RandomTwoQubitGates: RandomTwoQubitGatesArgs,
}

# AnastazData: Dict
# CUSTOM_ANSATZ_MAPPING : Ansatz, Args (class per ansatz)
# Args.parse_obj(**AnasatzData)



import pyomo
from pyomo import core

from classiq_interface.pyomo_extension import (
    inequality_expression,
    equality_expression,
    set_pprint,
)

pyomo.core.expr.logical_expr.InequalityExpression.getname = (
    inequality_expression.getname
)

pyomo.core.expr.logical_expr.EqualityExpression.getname = equality_expression.getname

pyomo.core.base.set.Set._pprint_members = set_pprint._pprint_members



from enum import Enum
from typing import Type, Union

import pydantic

from classiq_interface.generator import function_params


class AmplitudeEstimationInputs(Enum):
    IN = "IN"


AmplitudeEstimationOutputs = function_params.DefaultOutputNames


class AmplitudeEstimation(function_params.FunctionParams):
    """
    Creates a quantum circuit for ampitude estimation
    Provide the state preparation with a qasm string
    """

    state_preparation: str = pydantic.Field(
        description='The state preparation circuit in qasm format. Replace "..." with '
        "'...'. The total number of qubits is the sum of the state preparation qubits "
        "and `num_eval_qubits`"
    )

    objective_qubits: Union[
        pydantic.conlist(pydantic.conint(ge=0), min_items=1), pydantic.conint(ge=0)
    ] = pydantic.Field(
        default=0,
        description='The list of "good" qubits. The good states are the ones that have '
        "1's in the positions defined by objective qubits. The indices in this list "
        "must be in the range defined by `state_preparation`",
    )

    num_eval_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of qubits to evaluate on the amplitude estimation. "
        "More evaluation qubits provide a better estimate of the good states' amplitude"
    )

    _input_names: Type[Enum] = pydantic.PrivateAttr(default=AmplitudeEstimationInputs)
    _output_names: Type[Enum] = pydantic.PrivateAttr(default=AmplitudeEstimationOutputs)

    @pydantic.validator("state_preparation")
    def vscode_qasm2circuit(cls, v):
        VSCODE_QUOTE = "'"
        QASM_QUOTES = '"'
        return v.replace(VSCODE_QUOTE, QASM_QUOTES)



import enum
from typing import Union, Optional, Dict

import pydantic


class ExecutionStatus(str, enum.Enum):
    SUCCESS = "success"
    ERROR = "error"


class VaRResult(pydantic.BaseModel):
    var: float = None
    alpha: float = None


class FinanceSimulationResults(pydantic.BaseModel):
    var_results: Optional[VaRResult] = None
    result: Optional[float] = None


class GroverSimulationResults(pydantic.BaseModel):
    result: Dict[str, Union[float, int]]


class ExecutionResult(pydantic.BaseModel):
    status: ExecutionStatus
    details: Union[FinanceSimulationResults, GroverSimulationResults, str]



import random
import string
from typing import Optional, Dict, List

import pydantic

from classiq_interface.generator import function_param_list, function_params as f_params

_SUFFIX_LEN = 6
BAD_FUNCTION_ERROR_MSG = "Unknown function"
BAD_INPUT_ERROR_MSG = "Bad input name given"
BAD_OUTPUT_ERROR_MSG = "Bad output name given"


class FunctionCall(pydantic.BaseModel):
    function: pydantic.constr(min_length=1) = pydantic.Field(
        description="The function applied by the segment"
    )
    function_params: f_params.FunctionParams = pydantic.Field(
        description="The parameters necessary for defining the function"
    )
    inputs: Dict[
        pydantic.constr(min_length=1), pydantic.constr(min_length=1)
    ] = pydantic.Field(
        default_factory=dict,
        description="A mapping from the input name to the wire it connects to",
    )
    outputs: Dict[
        pydantic.constr(min_length=1), pydantic.constr(min_length=1)
    ] = pydantic.Field(
        default_factory=dict,
        description="A mapping from the output name to the wire it connects to",
    )
    add_as_single_gate: bool = pydantic.Field(
        default=True,
        description="Whether the segment is appended as a gate or "
        "composed into the circuit.",
    )

    name: Optional[pydantic.constr(min_length=1)] = pydantic.Field(
        default=None, description="The name of the segment. Determined automatically."
    )

    @pydantic.validator("name", always=True)
    def create_name(cls, name, values):
        function = values.get("function")
        if function is None:
            return name

        suffix = "".join(
            random.choice(string.ascii_letters + string.digits)
            for _ in range(_SUFFIX_LEN)
        )
        name = f"{function}_{suffix}"
        return name

    @pydantic.validator("function_params", pre=True)
    def parse_function_params(cls, function_params, values):
        if isinstance(function_params, pydantic.BaseModel):
            return function_params

        function = values.get("function")
        if function is None:
            return function_params

        func_class = [
            seg
            for seg in function_param_list.get_function_param_list()
            if seg.__name__ == function
        ]

        if not func_class:
            raise ValueError(f"{BAD_FUNCTION_ERROR_MSG}: {function}")

        return func_class[0].parse_obj(function_params)

    @pydantic.validator("inputs")
    def validate_input_names(cls, inputs, values):
        params = values.get("function_params")
        if params is None:
            return inputs

        invalid_names = FunctionCall._get_invalid_io_names(
            inputs.keys(), params, f_params.IO.Input
        )
        if invalid_names:
            raise ValueError(f"{BAD_INPUT_ERROR_MSG}: {invalid_names}")

        return inputs

    @pydantic.validator("outputs")
    def validate_output_names(cls, outputs, values):
        params = values.get("function_params")
        if params is None:
            return outputs

        invalid_names = FunctionCall._get_invalid_io_names(
            outputs.keys(), params, f_params.IO.Output
        )
        if invalid_names:
            raise ValueError(f"{BAD_OUTPUT_ERROR_MSG}: {invalid_names}")

        return outputs

    @staticmethod
    def _get_invalid_io_names(
        names: List[str], params: f_params.FunctionParams, io: f_params.IO
    ) -> List[str]:
        return [name for name in names if not params.is_valid_io_name(name, io)]



from enum import Enum
from typing import Type, List
from classiq_interface.generator.state_preparation import is_power_of_two
import numpy as np

import pydantic

from classiq_interface.generator.complex_type import Complex
from classiq_interface.generator.function_params import FunctionParams


class SparseAmpLoadOutputs(Enum):
    OUTPUT_STATE = "OUTPUT_STATE"


def amplitudes_sum_to_one(amp):
    if round(sum(abs(np.array(amp)) ** 2), 8) != 1:
        raise ValueError("Probabilities do not sum to 1")
    return amp


class SparseAmpLoad(FunctionParams):
    """
    loads a sparse amplitudes vector
    """

    num_qubits: pydantic.PositiveInt = pydantic.Field(
        description="The number of qubits in the circuit."
    )
    amplitudes: List[Complex] = pydantic.Field(description="amplitudes vector to load")

    _is_power_of_two = pydantic.validator("amplitudes", allow_reuse=True)(
        is_power_of_two
    )
    _is_sum_to_one = pydantic.validator("amplitudes", allow_reuse=True)(
        amplitudes_sum_to_one
    )

    _output_names: Type[Enum] = pydantic.PrivateAttr(default=SparseAmpLoadOutputs)



from enum import Enum
from typing import List, Type

import numpy as np
import pydantic

from classiq_interface.generator.complex_type import Complex
from classiq_interface.generator import function_params


class StatePropagatorInputs(Enum):
    IN = "IN"


StatePropagatorOutputs = function_params.DefaultOutputNames


class StatePropagator(function_params.FunctionParams):
    """
    Creates a quantum circuit that propagates the start state vector to the end state vector,
    both are assumed to be pure states.
    The default start state vector is |000...0>.
    """

    end_state_vector: List[Complex] = pydantic.Field(
        description="The desired state vector at the end of the operator."
        " Must be of size 2**num_qubits. Does not have to be "
        "normalized"
    )

    start_state_vector: List[Complex] = pydantic.Field(
        default_factory=list,
        description="The  state vector at the input of the operator."
        " Must be of size 2**num_qubits. Does not have to be"
        " normalized",
    )

    _input_names: Type[Enum] = pydantic.PrivateAttr(default=StatePropagatorInputs)
    _output_names: Type[Enum] = pydantic.PrivateAttr(default=StatePropagatorOutputs)

    @pydantic.validator("start_state_vector", always=True)
    def validate_start_state(cls, start_state_vector, values):
        end_state_vector = values.get("end_state_vector")
        num_qubits = int(np.log2(len(end_state_vector)))
        if len(start_state_vector) == 0:
            default_start_state_vector = [0.0 for _ in range(2 ** num_qubits)]
            default_start_state_vector[0] = 1.0
            start_state_vector = default_start_state_vector

        if len(start_state_vector) != len(end_state_vector):
            raise ValueError("Start and end state vectors are of non-equal length")

        return start_state_vector



from typing import Optional, List

import pydantic
from classiq_interface.generator import finance, grover_operator
from classiq_interface.hybrid import problem_input, encoding_type


class HybridMetadata(pydantic.BaseModel):
    problem: problem_input.ProblemInput
    encoding_type: encoding_type.EncodingType
    num_auxiliary: Optional[pydantic.conint(ge=0)] = 0
    permutation: List[int]
    is_qaoa: bool = False


class FinanceMetadata(pydantic.BaseModel):
    finance_attribute: Optional[finance.Finance]


class GroverMetadata(pydantic.BaseModel):
    grover_attribute: Optional[grover_operator.GroverOperator]


class GenerationMetadata(pydantic.BaseModel):
    hybrid: Optional[HybridMetadata]
    finance: Optional[FinanceMetadata]
    grover: Optional[GroverMetadata]



from typing import Optional, Union, List, Literal

from classiq_interface.hybrid.mht_input import (
    EntanglementStructureType,
    EntanglementBlocksType,
    RotationBlocksType,
)

import pydantic


class CustomAnsatzArgs(pydantic.BaseModel):
    num_qubits: int


class SeparateU3Args(CustomAnsatzArgs):
    pass


class HypercubeArgs(CustomAnsatzArgs):
    layer_count: int = 2


class EntanglingLayersArgs(CustomAnsatzArgs):
    layer_count: int = (2,)


class RandomArgs(CustomAnsatzArgs):
    gate_count: int = (100,)
    gate_probabilities: dict = ({"cx": 0.5, "u": 0.5},)
    random_seed: int = None


class RandomTwoQubitGatesArgs(CustomAnsatzArgs):
    random_two_qubit_gate_count_factor: float = (1.0,)
    random_seed: int = None


class TwoLocalArgs(CustomAnsatzArgs):
    rotation_blocks: Optional[
        Union[RotationBlocksType, List[RotationBlocksType]]
    ] = RotationBlocksType.ry
    entanglement_blocks: Optional[
        Union[EntanglementBlocksType, List[EntanglementBlocksType]]
    ] = EntanglementBlocksType.cx
    entanglement: EntanglementStructureType = EntanglementStructureType.full
    reps: int = 3



import tempfile

from pyomo.core import ConcreteModel

from classiq_interface.combinatorial_optimization import model_serializer, model_parser
import io


def pprint_string(model: ConcreteModel) -> str:
    f = io.StringIO()
    model.pprint(ostream=f)
    return f.getvalue()


def serialize_and_parse(model: ConcreteModel) -> ConcreteModel:
    serialization_dict = model_serializer.to_json(model, return_dict=True)
    model_copy, _ = model_parser.from_json(obj_dict=serialization_dict)

    model_string = pprint_string(model)
    model_copy_string = pprint_string(model_copy)

    assert model_string == model_copy_string, model_copy_string
    return model_copy



from typing import Optional

from classiq_interface.generator.arithmetic import ArithmeticOracle
from classiq_interface.generator.function_params import FunctionParams
from classiq_interface.generator.state_preparation import StatePreparation


class GroverOperator(FunctionParams):
    oracle: ArithmeticOracle
    state_preparation: Optional[StatePreparation] = None
    diffuser: Optional[str] = None

    def create_io_enums(self):
        self._input_names = self.oracle._input_names
        self._output_names = self.oracle._input_names

    class Config:
        extra = "forbid"



from typing import Type, Tuple

import pydantic

from classiq_interface.generator.problem_properties import (
    ProblemProperties,
)
from classiq_interface.generator.function_params import FunctionParams
from classiq_interface.hybrid.mht_qaoa_input import MhtQaoaInput
from classiq_interface.hybrid.problem_input import ProblemInput


class QaoaAnsatz(FunctionParams):
    """
    Parametric circuit for an optimization QAOA solver.
    """

    problem_properties: ProblemProperties = pydantic.Field(
        description="Properties of the " "combinatorial optimization " "problem"
    )

    @pydantic.validator("problem_properties")
    def validate_problem_supported_by_qaoa(cls, problem_properties):
        _SUPPORTED_PROBLEM_TYPES: Tuple[Type[ProblemInput], ...] = (MhtQaoaInput,)

        if not isinstance(problem_properties.problem, _SUPPORTED_PROBLEM_TYPES):
            raise ValueError(
                f"{problem_properties.problem.name} is not supported by QAOA"
            )

        return problem_properties



import pyomo.environ as pyo
import networkx as nx


def mis_pyomo_model(graph: nx.Graph) -> pyo.ConcreteModel:
    model = pyo.ConcreteModel()
    model.Nodes = pyo.Set(initialize=list(graph.nodes))
    model.Arcs = pyo.Set(initialize=list(graph.edges))

    model.x = pyo.Var(model.Nodes, domain=pyo.Binary)

    @model.Constraint(model.Arcs)
    def independent_rule(model, node1, node2):
        return model.x[node1] + model.x[node2] <= 1

    model.cost = pyo.Objective(
        expr=sum(model.x[node] for node in model.Nodes), sense=pyo.maximize
    )

    return model



import enum
from typing import Literal, Union, List, Optional, Any

import pydantic

from classiq_interface.hybrid.problem_input import ProblemInput, OptimizationProblemName


class RotationBlocksType(str, enum.Enum):
    rx = "rx"
    ry = "ry"
    rz = "rz"


class EntanglementBlocksType(str, enum.Enum):
    cx = "cx"
    cy = "cy"
    cz = "cz"


class EntanglementStructureType(str, enum.Enum):
    linear = "linear"
    full = "full"
    circular = "circular"
    sca = "sca"


class MHTInput(ProblemInput):
    name: Literal[OptimizationProblemName.MHT] = pydantic.Field(
        default=OptimizationProblemName.MHT, description="Name of optimization problem."
    )
    plot_count: pydantic.conint(gt=0, lt=20) = pydantic.Field(
        default=3, description="Number of plots."
    )
    max_track_count: pydantic.conint(gt=0, lt=20) = pydantic.Field(
        default=2, description="Maximum number of tracks."
    )
    penalty_energy: Union[int, float] = pydantic.Field(
        default=10,
        description="Penalty energy for invalid solutions, "
        "The value effects the converges rate."
        " Small positive values are preferred",
    )
    plot_track_matching: Optional[List[int]] = pydantic.Field(
        default=None, description="known matching between plots and tracks"
    )
    reps: int = pydantic.Field(
        default=3,
        description="Amount of repetitions of a rotation layer and entanglement layer.",
    )
    rotation_blocks: Union[
        RotationBlocksType, List[RotationBlocksType]
    ] = pydantic.Field(
        default=RotationBlocksType.ry,
        description="The gates used in the rotation layer.",
    )
    entanglement_blocks: Union[
        EntanglementBlocksType, List[EntanglementBlocksType]
    ] = pydantic.Field(
        default=EntanglementBlocksType.cx,
        description="The gates used in the entanglement layer.",
    )
    entanglement: EntanglementStructureType = pydantic.Field(
        default=EntanglementStructureType.full,
        description="Structure of entanglement layer.",
    )
    data_for_cost_server: Any = pydantic.Field(
        default=None,
        description="Additional data to forward to the external"
        " cost server, e.g. configuration",
    )

    @pydantic.validator("plot_track_matching", always=True)
    def init_plot_track_matching(cls, plot_track_matching, values):
        plot_count = values.get("plot_count")
        if plot_count is None:
            raise ValueError("Plot count is mandatory")
        if plot_track_matching is None:
            plot_track_matching = [0] * plot_count

        return plot_track_matching

    def is_valid_cost(self, cost: float) -> bool:
        return True



from typing import List
import numpy as np


def get_rand_adjacency_matrix(num_vertices: int) -> List[List[bool]]:
    adjacency_matrix = np.random.choice([0, 1], size=(num_vertices, num_vertices))
    for i in range(num_vertices):
        for j in range(num_vertices):
            if i > j:
                adjacency_matrix[i, j] = adjacency_matrix[j, i]
    adjacency_matrix -= np.diag(np.diag(adjacency_matrix))
    return adjacency_matrix.tolist()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Functions for loading Pyomo objects from json
"""
import ast
import gzip
import inspect
import json
import time
import weakref

from classiq_interface.pyomo_extension import pyomo
from pyomo.core.base import _GeneralVarData
from pyomo.core.expr.numeric_expr import ExpressionBase
from pyomo.environ import Set, RangeSet, Suffix, Objective

from classiq_interface.combinatorial_optimization.model_io_comon import (
    StoreSpec,
    may_have_subcomponents,
)

__format_version__ = 1


def get_last_object_from_lookup(lookup, object_name):
    new_object = None
    for object_lookup in reversed(lookup.values()):
        if hasattr(object_lookup, "name") and object_name == object_lookup.name:
            new_object = object_lookup
            break
    return new_object


PYOMO_CLASS_MAPPING = {
    "<class 'pyomo.core.base.PyomoModel.ConcreteModel'>": pyomo.core.base.PyomoModel.ConcreteModel,
    "<class 'pyomo.core.base.param.ScalarParam'>": pyomo.core.base.param.ScalarParam,
    "<class 'pyomo.core.base.var.IndexedVar'>": pyomo.core.base.var.IndexedVar,
    "<class 'pyomo.core.base.var._GeneralVarData'>": pyomo.core.base.var._GeneralVarData,
    "<class 'pyomo.core.base.logical_constraint.IndexedLogicalConstraint'>": pyomo.core.base.logical_constraint.IndexedLogicalConstraint,
    "<class 'pyomo.core.base.logical_constraint._GeneralLogicalConstraintData'>": pyomo.core.base.logical_constraint._GeneralLogicalConstraintData,
    "<class 'pyomo.core.base.constraint.IndexedConstraint'>": pyomo.core.base.constraint.IndexedConstraint,
    "<class 'pyomo.core.base.constraint._GeneralConstraintData'>": pyomo.core.base.constraint._GeneralConstraintData,
    "<class 'pyomo.core.base.constraint.ScalarConstraint'>": pyomo.core.base.constraint.ScalarConstraint,
    "<class 'pyomo.core.expr.numvalue.NumericConstant'>": pyomo.core.expr.numvalue.NumericConstant,
    "<class 'pyomo.core.base.objective.ScalarObjective'>": pyomo.core.base.objective.ScalarObjective,
    "<class 'pyomo.core.expr.numeric_expr.SumExpression'>": pyomo.core.expr.numeric_expr.SumExpression,
    "<class 'pyomo.core.expr.numeric_expr.ProductExpression'>": pyomo.core.expr.numeric_expr.ProductExpression,
    "<class 'pyomo.core.expr.logical_expr.InequalityExpression'>": pyomo.core.expr.logical_expr.InequalityExpression,
    "<class 'pyomo.core.expr.logical_expr.EqualityExpression'>": pyomo.core.expr.logical_expr.EqualityExpression,
    "<class 'pyomo.core.base.set.OrderedScalarSet'>": pyomo.core.base.set.OrderedScalarSet,
    "<class 'pyomo.core.base.set.FiniteScalarRangeSet'>": pyomo.core.base.set.FiniteScalarRangeSet,
    "<class 'pyomo.core.expr.numeric_expr.MonomialTermExpression'>": pyomo.core.expr.numeric_expr.MonomialTermExpression,
    "<class 'pyomo.core.base.set.FiniteScalarSet'>": pyomo.core.base.set.FiniteScalarSet,
    "<class 'float'>": float,
    "<class 'int'>": int,
    "<class 'tuple'>": tuple,
}


def _read_component(
    obj_dict: dict,
    store_spec: StoreSpec,
    lookup: dict = None,
    suffixes: dict = None,
    root_name: str = None,
):
    """
    Read a component dictionary into a model
    """
    if lookup is None:
        lookup = {}
    if suffixes is None:
        suffixes = {}

    try:
        root_dict = obj_dict[root_name]
    except KeyError as e:
        if store_spec.ignore_missing:
            return
        else:
            raise e

    # read sub data and overwrite
    components_dict = _read_component_data(
        root_dict["data"], store_spec, lookup=lookup, suffixes=suffixes
    )
    for object_name in components_dict.keys():
        object_lookup = get_last_object_from_lookup(
            lookup=lookup, object_name=root_name
        )
        if object_lookup is not None:
            components_dict[object_name] = object_lookup

    # create obj from dict
    pyomo_class = PYOMO_CLASS_MAPPING[root_dict["__type__"]]
    if len(components_dict) == 1:
        obj = list(components_dict.values())[0]
        if hasattr(obj, "construct") and isinstance(obj, RangeSet):
            obj._constructed = False
            obj.construct()
    else:
        index_name = root_dict.get("index", None)
        if index_name in {"UnindexedComponent_set", None}:
            args_parsed = [ast.literal_eval(arg) for arg in root_dict["data"]]
        else:
            args_parsed = get_last_object_from_lookup(lookup, index_name)

        if issubclass(pyomo_class, Set):
            obj = pyomo_class(initialize=args_parsed)
        else:
            obj = pyomo_class(args_parsed)
            if hasattr(obj, "construct"):
                obj.construct()

            for key, el in components_dict.items():
                el._component = weakref.ref(obj)
                obj._data[key] = el

    # set attributes to obj
    attr_list, filter_function = store_spec.get_class_attr_list(pyomo_class)
    if filter_function is not None:
        attr_list = filter_function(obj, root_dict)
    obj = _set_attributes_from_dict(obj, root_dict, attr_list, store_spec)

    # make a dict of suffixes to read at the end
    if isinstance(obj, Suffix):
        if store_spec.include_suffix:
            if (
                store_spec.suffix_filter is None
                or root_name in store_spec.suffix_filter
            ):
                suffixes[root_dict["__id__"]] = root_dict["data"]  # is populated

    lookup[root_dict["__id__"]] = obj
    return obj


def _read_component_data(
    obj_dict: dict, store_spec: StoreSpec, lookup: dict = None, suffixes: dict = None
) -> dict:
    """
    Read a Pyomo component json_str data in from a dict.
    Args:
        obj_dict: dictionary to read from
        store_spec: StoreSpec object specifying what to read in
        lookup: a lookup table for id to component for reading suffixes
        suffixes: a list of suffixes put off reading until end
    Returns:
        Dict
    """
    if lookup is None:
        lookup = {}
    if suffixes is None:
        suffixes = {}

    attr_list = None
    filter_function = None

    pyomo_obj_mapping = {}
    for key, element_dict in obj_dict.items():

        pyomo_class = PYOMO_CLASS_MAPPING[element_dict["__type__"]]

        if attr_list is None:  # assume all items are same
            attr_list, filter_function = store_spec.get_data_class_attr_list(
                pyomo_class
            )

        initialization_dict = {}
        if "expr" in element_dict:
            expr_name, odict = next(iter(element_dict["expr"].items()))
            initialization_dict["expr"] = _read_component(
                {expr_name: odict},
                store_spec,
                lookup=lookup,
                suffixes=suffixes,
                root_name=expr_name,
            )

        elif "args" in element_dict:
            initialization_dict["args"] = [
                _read_component(
                    odict,
                    store_spec,
                    lookup=lookup,
                    suffixes=suffixes,
                    root_name=list(odict.keys())[0],
                )
                for odict in element_dict["args"].values()
            ]

        class_init_params = inspect.signature(pyomo_class.__init__).parameters
        for attr in attr_list:  # read in desired attributes
            if attr in class_init_params:
                if attr == "domain":
                    element_dict[attr] = getattr(pyomo.environ, element_dict[attr])
                initialization_dict[attr] = element_dict[attr]

        pyomo_obj = pyomo_class(**initialization_dict)

        if "__pyomo_components__" in element_dict:
            # read sub-components of block-like
            for component_name, odict in element_dict["__pyomo_components__"].items():
                component = _read_component(
                    {component_name: odict},
                    store_spec,
                    lookup=lookup,
                    suffixes=suffixes,
                    root_name=component_name,
                )
                if isinstance(component_name, Objective) and hasattr(
                    component, "construct"
                ):
                    component._constructed = False
                    component.construct()

                setattr(pyomo_obj, component_name, component)

        if filter_function:
            attr_list = filter_function(pyomo_obj, element_dict)

        pyomo_obj = _set_attributes_from_dict(
            pyomo_obj, element_dict, attr_list, store_spec
        )

        lookup[element_dict["__id__"]] = pyomo_obj
        pyomo_obj_mapping[ast.literal_eval(key)] = pyomo_obj
    return pyomo_obj_mapping


def _set_attributes_from_dict(pyomo_obj, element_dict, attr_list, store_spec) -> object:
    for attr in attr_list:  # read in desired attributes
        if attr not in element_dict:
            if store_spec.ignore_missing:
                continue
            else:
                raise KeyError

        if attr in store_spec.set_functions:
            store_spec.set_functions[attr](pyomo_obj, element_dict[attr])
        else:
            setattr(pyomo_obj, attr, element_dict[attr])
        if isinstance(pyomo_obj, (float, int)) and attr == "value":
            pyomo_obj = element_dict[attr]
    return pyomo_obj


def component_data_from_dict(obj_dict, obj, store_spec: StoreSpec):
    """
    Component data to a dict.
    """
    attr_list, filter_function = store_spec.get_data_class_attr_list(obj)
    if filter_function is not None:
        attr_list = filter_function(obj, obj_dict)

    obj = _set_attributes_from_dict(obj, obj_dict, attr_list, store_spec)

    if may_have_subcomponents(obj):  # read sub-components of block-like
        for _ in obj.component_objects(descend_into=False):
            _read_component(obj_dict["__pyomo_components__"], store_spec)


def _read_suffixes(lookup: dict, suffixes: dict):
    """
    Go through the list of suffixes and read the data back in.
    Args:
        lookup: a lookup table to go from id to component
        suffixes: a dictionary with suffix id keys and value dict value
    Returns:
        None
    """
    for uid in suffixes:
        keys = suffixes[uid]
        s = lookup[uid]  # suffixes keys are ids, so get suffix component
        for key in keys:  # set values from value dict
            try:
                kc = lookup[int(key)]  # use int because json turn keys to string
            except KeyError:
                continue
            s[kc] = keys[key]


def from_json(
    obj_dict=None,
    file_name=None,
    json_str=None,
    store_spec=None,
    gz=None,
    root_name=None,
):
    """
    Load the state of a Pyomo component state from a dictionary, json file, or
    json string.  Must only specify one of obj_dict, file_name, or s as a non-None value.
    This works by going through the model and loading the state of each
    sub-component of obj. If the saved state contains extra information, it is
    ignored.  If the save state doesn't contain an entry for a model component
    that is to be loaded an error will be raised, unless ignore_missing = True.
    Args:
        obj_dict: State dictionary to load, if None, check file_name and json_str
        file_name: JSON file to load, only used if obj_dict is None
        json_str: JSON string to load only used if both obj_dict and file_name are None
        store_spec: StoreSpec object specifying what to load
        gz: If True assume the file specified by file_name is gzipped. The default is
            True if file_name ends with '.gz' otherwise False.
    Returns:
        Dictionary with some performance information. The keys are
        "load_file_time", how long in seconds it took to load the json file
        "read_dict_time", how long in seconds it took to read models state
        "read_suffixes_time", how long in seconds it took to read suffixes
    """
    start_time = time.time()

    if gz is None:
        if isinstance(file_name, str):
            gz = file_name.endswith(".gz")
        else:
            gz = False

    if obj_dict is not None:  # Existing Python dict (for in-memory stuff).
        pass
    elif file_name is not None:
        if gz:
            with gzip.open(file_name, "r") as f:
                fr = f.read()
                obj_dict = json.loads(fr)
        else:
            with open(file_name, "r") as f:
                obj_dict = json.load(f)
    elif json_str is not None:
        obj_dict = json.loads(json_str)
    else:
        raise Exception("Need to specify a data source to load from")
    load_time = time.time()

    if root_name is None:
        for key in obj_dict:
            if key.startswith("__") and key.endswith("__"):
                continue  # This is metadata or maybe some similar future addition.
            else:
                root_name = key
                break  # should be one root

    store_spec = store_spec if store_spec else StoreSpec()
    lookup = {}
    suffixes = {}
    obj = _read_component(
        obj_dict, store_spec, lookup=lookup, suffixes=suffixes, root_name=root_name
    )
    read_time = time.time()

    _read_suffixes(lookup, suffixes)
    suffix_time = time.time()

    performance_dict = {
        "load_file_time": load_time - start_time,
        "read_dict_time": read_time - load_time,
        "read_suffixes_time": suffix_time - read_time,
    }
    return obj, performance_dict



from typing import List, Literal
from typing import Optional

import numpy as np
import pydantic

from classiq_interface.generator import distance
from classiq_interface.hybrid.problem_input import ProblemInput, OptimizationProblemName


class TSPInput(ProblemInput):
    name: Literal[OptimizationProblemName.TSP] = pydantic.Field(
        default=OptimizationProblemName.TSP, description="Name of optimization problem."
    )
    num_cities: pydantic.conint(gt=1) = pydantic.Field(
        default=..., description="Number of cities in the TSP problem."
    )
    distance_matrix: Optional[List[List[float]]] = pydantic.Field(
        default=None, description="Matrix consisting of distance between the cities."
    )

    @pydantic.root_validator
    def init_distance_matrix(cls, values):
        num_cities, distance_matrix = values.get("num_cities"), values.get(
            "distance_matrix"
        )
        if num_cities is None:
            raise ValueError("Num cities is mandatory")
        if distance_matrix is None:
            values["distance_matrix"] = distance.get_rand_distance_matrix(
                num_points=num_cities
            )

        return values

    @pydantic.validator("distance_matrix")
    def check_distance_matrix(cls, matrix):
        array = np.array(matrix)
        if not array.shape[0] == array.shape[1]:
            raise ValueError("distance matrix must be square")
        if not np.all(array >= 0):
            raise ValueError("All the elements of distance matrix must be non-negative")
        if not np.allclose(array, array.T):
            raise ValueError("Distance Matrix is not symmetric")
        return matrix

    def is_valid_cost(self, cost: float) -> bool:
        distance_matrix = np.array(self.distance_matrix)
        num_cities = distance_matrix.shape[0]
        max_distance = np.amax(distance_matrix)
        max_cost = max_distance * (num_cities - 1)
        _MIN_COST = 0

        return _MIN_COST < cost and approx_lte(x=cost, y=max_cost)


def approx_lte(x, y) -> bool:
    return x < y or np.isclose(x, y)



from typing import List, Optional, Dict, Any

import pydantic

from classiq_interface.generator.mcmt_method import McmtMethod
from classiq_interface.generator import custom_ansatz
from classiq_interface.generator.problem_properties import (
    ProblemProperties,
)
from classiq_interface.generator.function_params import FunctionParams
from classiq_interface.mixer import MixerType, MIXER_TYPES_DEFAULT
from classiq_interface.hybrid.encoding_type import EncodingType


class VQEAnsatz(FunctionParams):
    """
    Parametric circuit for an optimization VQE solver.
    """

    problem_properties: ProblemProperties = pydantic.Field(
        description="Properties of the " "combinatorial optimization " "problem"
    )
    mixer_types: List[MixerType] = pydantic.Field(
        default_factory=list,
        description="Ordered list of mixers applied in the ansatz.",
    )

    mcmt_method: McmtMethod = pydantic.Field(
        default=McmtMethod.vchain,
        description="multi controlled gates decomposition method.",
    )
    custom_ansatz_name: Optional[custom_ansatz.CustomAnsatzType] = pydantic.Field(
        default=None, description="A custom ansatz type"
    )
    custom_ansatz_args: Optional[Dict[str, Any]] = pydantic.Field(
        default=None, description="the arguments to the custom ansatz"
    )

    @pydantic.validator("mixer_types", always=True)
    def check_mixer_types(cls, mixer_types, values):
        if mixer_types is None:
            mixer_types = MIXER_TYPES_DEFAULT[type(values["problem"])]
        return mixer_types

    @pydantic.validator("mcmt_method", always=True)
    def check_mcmt_method(cls, mcmt_method, values):
        if mcmt_method is None:
            if values["encoding_type"] == EncodingType.SERIAL:
                return McmtMethod.vchain
            else:  # == EncodingType.NODES
                return McmtMethod.standard
        return mcmt_method

    @property
    def custom_ansatz_args_class(self):
        if self.custom_ansatz_name is None:
            return None
        return custom_ansatz.CUSTOM_ANSATZ_ARGS_MAPPING[self.custom_ansatz_name]



from typing import Union

import pydantic
from classiq_interface.finance.function_input import FinanceFunctionInput
from classiq_interface.finance.model_input import FinanceModelInput
from classiq_interface.generator import function_params


class Finance(function_params.FunctionParams):
    model: FinanceModelInput = pydantic.Field(description="Load a financial model")
    finance_function: Union[FinanceFunctionInput] = pydantic.Field(
        description="The finance function to solve the model"
    )



from typing import Literal, Optional
import pydantic

from classiq_interface.backend.quantum_backend_providers import (
    ProviderVendor,
    BACKEND_PROVIDERS_DICT,
)


class BackendPreferences(pydantic.BaseModel):
    backend_service_provider: str = pydantic.Field(
        description="Provider company for the requested backend."
    )
    backend_name: str = pydantic.Field(description="Name of the requested backend.")


class AwsBackendPreferences(BackendPreferences):
    _BACKEND_SERVICE_PROVIDER = ProviderVendor.AWS_BRAKET
    backend_service_provider: Literal[_BACKEND_SERVICE_PROVIDER.value]
    backend_name: Literal[
        tuple(name.value for name in BACKEND_PROVIDERS_DICT[_BACKEND_SERVICE_PROVIDER])
    ]
    aws_access_key_id: Optional[str] = pydantic.Field(description="AWS Access Key ID")
    aws_secret_access_key: Optional[str] = pydantic.Field(
        description="AWS Secret Access Key"
    )
    region_name: Optional[str] = pydantic.Field(description="AWS Region name")
    s3_bucket_name: str = pydantic.Field(description="S3 Bucket Name")
    s3_bucket_key: str = pydantic.Field(description="S3 Bucket Key")


class IBMBackendPreferences(BackendPreferences):
    _BACKEND_SERVICE_PROVIDER = ProviderVendor.IBMQ
    backend_service_provider: Literal[_BACKEND_SERVICE_PROVIDER.value]
    backend_name: Literal[
        tuple(name.value for name in BACKEND_PROVIDERS_DICT[_BACKEND_SERVICE_PROVIDER])
    ]


class AzureBackendPreferences(BackendPreferences):
    _BACKEND_SERVICE_PROVIDER = ProviderVendor.AZURE_QUANTUM
    backend_service_provider: Literal[_BACKEND_SERVICE_PROVIDER.value]
    backend_name: Literal[
        tuple(name.value for name in BACKEND_PROVIDERS_DICT[_BACKEND_SERVICE_PROVIDER])
    ]



from enum import Enum
from typing import Optional, List, Union

import pydantic
from classiq_interface.generator.custom_gates import ValidGateName


class UnrollerType(str, Enum):
    THREE_QUBIT_GATES = "three_qubit_gates"


class TranspilationPreferences(pydantic.BaseModel):
    """
    Preferences for running transpiler on the generated quantum circuit.
    """

    is_sub_circuit: bool = pydantic.Field(
        default=False,
        description="Whether the transpiled circuit is part of a bigger circuit. "
        "This enables optimizations with assumptions on the start and end conditions of the circuit.",
    )
    unroller: Optional[Union[UnrollerType, List[ValidGateName]]] = pydantic.Field(
        default=None,
        description=f"Unroll {UnrollerType.THREE_QUBIT_GATES} or unroll to specific basis gates.",
    )

    @pydantic.validator("unroller")
    def validate_unroller(cls, unroller):
        if isinstance(unroller, list) and not unroller:
            raise ValueError("Empty basis gates is forbidden.")

        return unroller



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['classiq_interface',
 'classiq_interface.analyzer',
 'classiq_interface.backend',
 'classiq_interface.combinatorial_optimization',
 'classiq_interface.combinatorial_optimization.examples',
 'classiq_interface.execute',
 'classiq_interface.finance',
 'classiq_interface.generator',
 'classiq_interface.generator.preferences',
 'classiq_interface.generator.validations',
 'classiq_interface.hybrid',
 'classiq_interface.pyomo_extension',
 'classiq_interface.server']

package_data = \
{'': ['*']}

install_requires = \
['Pyomo>=6.0,<6.1',
 'more-itertools>=8.8.0,<9.0.0',
 'networkx>=2.5.1,<3.0.0',
 'numpy>=1.20.1,<2.0.0',
 'pillow>=8.3.1,<9.0.0',
 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'classiq-interface',
    'version': '0.1.2b0',
    'description': 'Classiq Interface',
    'long_description': 'See [classiq package](https://pypi.org/project/classiq/) README.\n\n### License\nSee [license](https://classiq.io/license).\n',
    'author': 'Classiq Technologies',
    'author_email': 'support@classiq.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://classiq.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from math import pi

_DEG_TO_RAD_RATIO = pi / 180


def phase_to_rad(deg: int) -> float:
    phase_in_rad = round(deg * _DEG_TO_RAD_RATIO, 4)
    return phase_in_rad


def to_canonical_phase(phase: float) -> float:
    return phase % (2 * pi)


def rad_to_deg(phase: float) -> int:
    return int(round(to_canonical_phase(phase) * (180 / pi)))



from enum import Enum


class ProviderVendor(Enum):
    IBMQ = "IBMQ"
    AZURE_QUANTUM = "Azure Quantum"
    AWS_BRAKET = "AWS Braket"


class BackendNames(Enum):
    AWS_BRAKET_SV1 = "SV1"
    AWS_BRAKET_TN1 = "TN1"
    AWS_BRAKET_DM1 = "dm1"
    AWS_BRAKET_ASPEN_9 = "Aspen-9"
    AWS_BRAKET_IONQ = "IonQ Device"
    IBMQ_AER_SIMULATOR = "aer_simulator"
    IBMQ_AER_SIMULATOR_STATEVECTOR = "aer_simulator_statevector"
    IBMQ_AER_SIMULATOR_DENSITY_MATRIX = "aer_simulator_density_matrix"
    IBMQ_AER_SIMULATOR_MATRIX_PRODUCT_STATE = "aer_simulator_matrix_product_state"
    AZURE_QUANTUM_IONQ = "IonQ"
    AZURE_QUANTUM_HONEYWELL = "Honeywell"


BACKEND_PROVIDERS_DICT = {
    ProviderVendor.AWS_BRAKET: [
        BackendNames.AWS_BRAKET_SV1,
        BackendNames.AWS_BRAKET_TN1,
        BackendNames.AWS_BRAKET_DM1,
        BackendNames.AWS_BRAKET_ASPEN_9,
        BackendNames.AWS_BRAKET_IONQ,
    ],
    ProviderVendor.IBMQ: [
        BackendNames.IBMQ_AER_SIMULATOR,
        BackendNames.IBMQ_AER_SIMULATOR_STATEVECTOR,
        BackendNames.IBMQ_AER_SIMULATOR_DENSITY_MATRIX,
        BackendNames.IBMQ_AER_SIMULATOR_MATRIX_PRODUCT_STATE,
    ],
    ProviderVendor.AZURE_QUANTUM: [
        BackendNames.AZURE_QUANTUM_IONQ,
        BackendNames.AZURE_QUANTUM_HONEYWELL,
    ],
}



import pydantic
from classiq_interface.generator.circuit_outline import Qubit
from classiq_interface.generator.slides import Slide, PhaseSlide


class CXSlideEntangler(Slide):
    """
    A slide of CX gates
    """

    # TODO: Replace to valid qubit.
    slide_first_control: Qubit = pydantic.Field(
        default=..., description="The control qubit of the first CX gate in the slide"
    )

    @pydantic.validator("slide_first_control")
    def first_control_validator(cls, first_control, values):
        targets = values.get("slide_targets")

        if targets is not None and first_control in targets:
            # TODO: Find out whether it is true.
            raise ValueError("first control can not be a target in the slide")

        return first_control


class PhaseCXSlideEntangler(PhaseSlide, CXSlideEntangler):
    pass


class SquareClusterEntanglerParameters(pydantic.BaseModel):
    num_of_qubits: pydantic.conint(ge=2)
    schmidt_rank: pydantic.conint(ge=0)


class Open2DClusterEntanglerParameters(pydantic.BaseModel):
    qubit_count: pydantic.conint(ge=2)
    schmidt_rank: pydantic.conint(ge=0)



from enum import Enum

import pydantic


class StatePrepOptimizationMethod(str, Enum):
    KL = "KL"
    L2 = "L2"
    L1 = "L1"
    MAX_PROBABILITY = "MAX_PROBABILITY"
    RANDOM = "RANDOM"


class OptimizationType(str, Enum):
    DEPTH = "depth"
    TWO_QUBIT_GATES = "two_qubit_gates"


class Optimization(pydantic.BaseModel):
    approximation_error: pydantic.confloat(ge=0, lt=1) = 0
    optimization_type: OptimizationType = OptimizationType.DEPTH



import enum
from typing import Optional, Union

import pydantic


class KnownFunctions(str, enum.Enum):
    VAR = "var"
    SHORTFALL = "expected short fall"
    X_SQUARE = "x**2"
    EUROPEAN_CALL_OPTION = "european call option"


class FunctionCondition(pydantic.BaseModel):
    threshold: float
    larger: bool = pydantic.Field(
        default=False,
        description="When true, function is set when input is larger to threshold and otherwise 0. Default is False.",
    )


class FinanceFunctionInput(pydantic.BaseModel):
    f: Union[str, KnownFunctions] = pydantic.Field(
        description="A callable function to solve the model"
    )
    variable: Optional[str] = pydantic.Field(
        default="x", description="Variable/s of the function"
    )
    condition: Optional[FunctionCondition] = pydantic.Field(
        description="The condition for the function"
    )
    polynomial_degree: Optional[int] = pydantic.Field(
        default=None,
        description="The polynomial degree of approximation, uses linear approximation by default",
    )
    use_chebyshev_polynomial_approximation: bool = pydantic.Field(
        default=False,
        description="Flag if to use chebyshev polynomial approximation for target function",
    )

    tail_probability: Optional[pydantic.confloat(gt=0, le=1)] = pydantic.Field(
        default=None,
        description="The required probability on the tail of the distribution (1 - percentile)",
    )

    class Config:
        extra = "forbid"

    @pydantic.validator("use_chebyshev_polynomial_approximation")
    def validate_polynomial_flag(cls, use_chebyshev_flag, values):
        polynomial_degree = values.get("polynomial_degree")
        if (use_chebyshev_flag is False and polynomial_degree is None) or (
            use_chebyshev_flag is True and polynomial_degree is not None
        ):
            return use_chebyshev_flag
        else:
            raise ValueError(
                "Degree must be positive and use_chebyshev_polynomial_approximation set to True"
            )

    @pydantic.validator("tail_probability", always=True)
    def validate_tail_probability_assignment_for_shortfall(
        cls, tail_probability, values
    ):

        if values.get("f") == KnownFunctions.SHORTFALL and not tail_probability:
            raise ValueError("Tail probability must be set for expected short fall")
        return tail_probability



from enum import Enum
from typing import Union, Set, List, Optional

from pydantic import BaseModel, Field, conint

Match = List[List[int]]


class AnalysisStatus(str, Enum):
    NONE = "none"
    SUCCESS = "success"
    CANCELLED = "cancelled"
    ERROR = "error"


class BasisGates(str, Enum):
    CX = "cx"
    CY = "cy"
    CZ = "cz"
    U = "u"
    U2 = "u2"
    P = "p"


class QuantumCircuitProperties(BaseModel):
    depth: conint(ge=0) = Field(default=..., description="Circuit depth")
    auxiliary_qubits: conint(ge=0) = Field(
        default=..., description="Number of Auxiliary qubits"
    )
    classical_bits: conint(ge=0) = Field(
        default=..., description="Number of classical bits"
    )
    gates_count: conint(ge=0) = Field(
        default=..., description="Total number of gates in the circuit"
    )
    multi_qubit_gates_count: conint(ge=0) = Field(
        default=..., description="Number of multi-qubit gates in circuit"
    )
    non_entangled_subcircuits_count: conint(ge=0) = Field(
        default=..., description="Number of non-entangled sub-circuit "
    )


class NativeQuantumCircuitProperties(QuantumCircuitProperties):
    native_gates: Set[BasisGates] = Field(
        default=..., description="Native gates used for decomposition"
    )


class PatternRecognitionResult(BaseModel):
    found_patterns: List[str]


class PatternMatchingResult(BaseModel):
    found_patterns: List[str]


class Circuit(BaseModel):
    closed_circuit_qasm: str
    image: str


class PatternAnalysis(BaseModel):
    pattern_matching: Optional[PatternMatchingResult] = Field(
        default=..., description="Pattern matching algorithm"
    )
    pattern_recognition: Optional[PatternRecognitionResult] = Field(
        default=..., description="Find unknown patterns"
    )
    circuit: Circuit = Field(
        default=..., description="Quantum circuit after pattern analysis"
    )


class Analysis(BaseModel):
    input_properties: QuantumCircuitProperties = Field(
        default=..., description="Input circuit properties"
    )
    native_properties: NativeQuantumCircuitProperties = Field(
        default=..., description="Transpiled circuit properties"
    )
    pattern_analysis: Optional[PatternAnalysis] = Field(
        default=None,
        description="Pattern analysis, including pattern matching and pattern recognition",
    )


class AnalysisResult(BaseModel):
    status: AnalysisStatus
    details: Union[Analysis, str]



from collections import defaultdict
from dataclasses import dataclass
from typing import List, Optional, Dict
from itertools import chain
import networkx as nx
import pydantic
from classiq_interface.generator.segments import FunctionCall

IO_MULTI_USE_ERROR_MSG = "Input and output names can only be used once"
IO_NAME_MISMATCH_ERROR_MSG = "Inputs and outputs are not identical"


@dataclass
class Wire:
    start: Optional[pydantic.constr(min_length=1)] = None
    end: Optional[pydantic.constr(min_length=1)] = None


def _parse_segment_inputs(segment: FunctionCall, wires: Dict[str, Wire]) -> None:
    if not segment.inputs:
        return

    for wire_name in segment.inputs.values():
        wire = wires[wire_name]

        if wire.end:
            raise ValueError(
                IO_MULTI_USE_ERROR_MSG
                + f". The name {wire_name} is used multiple times."
            )
        wire.end = segment.name


def _parse_segment_outputs(segment: FunctionCall, wires: Dict[str, Wire]) -> None:
    if not segment.outputs:
        return

    for wire_name in segment.outputs.values():
        wire = wires[wire_name]

        if wire.start:
            raise ValueError(
                IO_MULTI_USE_ERROR_MSG
                + f". The name {wire_name} is used multiple times."
            )
        wire.start = segment.name


def create_flow_graph(segment_list: List[FunctionCall]) -> nx.DiGraph:
    input_names = sorted(
        chain(*[segment.inputs.values() for segment in segment_list if segment.inputs])
    )
    output_names = sorted(
        chain(
            *[segment.outputs.values() for segment in segment_list if segment.outputs]
        )
    )

    if not input_names == output_names:
        raise ValueError(IO_NAME_MISMATCH_ERROR_MSG)

    wires = defaultdict(Wire)
    for segment in segment_list:
        _parse_segment_inputs(segment, wires)
        _parse_segment_outputs(segment, wires)

    edges = [(wire.start, wire.end) for wire in wires.values()]

    graph = nx.DiGraph()

    graph.add_nodes_from(
        [(segment.name, {"data": segment}) for segment in segment_list]
    )
    graph.add_edges_from(edges)

    return graph



import abc
from enum import Enum
from typing import Optional, ClassVar, Type, Any

import pydantic


class DefaultInputNames(Enum):
    pass


class DefaultOutputNames(Enum):
    OUT = "OUT"


class IO(Enum):
    Input = "Input"
    Output = "Output"


class FunctionParams(pydantic.BaseModel, abc.ABC):
    _input_names: Type[Enum] = pydantic.PrivateAttr(default=DefaultInputNames)
    _output_names: Type[Enum] = pydantic.PrivateAttr(default=DefaultOutputNames)

    def __init__(self, **data: Any):
        super().__init__(**data)
        self.create_io_enums()

    def get_io_enum(self, io: IO) -> Type[Enum]:
        if io == IO.Input:
            return self._input_names
        if io == IO.Output:
            return self._output_names

    def create_io_enums(self):
        pass

    def is_valid_io_name(self, name: str, io: IO) -> bool:
        return name in self.get_io_enum(io).__members__



import pydantic
from classiq_interface.generator.circuit_outline import Qubit


class Teleport(pydantic.BaseModel):
    qubit_data: Qubit
    qubit_transmitter: Qubit
    qubit_receiver: Qubit
    num_cifs = 2



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Functions for saving and loading Pyomo objects to json
"""
from pyomo.core.base import _GeneralVarData
from pyomo.core.base.indexed_component import IndexedComponent
from pyomo.core.expr.numeric_expr import ExpressionBase
from pyomo.network import Port
from pyomo.environ import (
    value,
    Var,
    ExternalFunction,
    Expression,
    RangeSet,
    Suffix,
    Param,
    Component,
)
import pyomo


def may_have_subcomponents(obj):
    """
    Args:
        obj: an object.
    Returns:
        True if the object has a callable component_objects method, otherwise
        False.
    """
    if hasattr(obj, "component_objects"):
        if hasattr(obj.component_objects, "__call__"):
            return True


def _set_active(obj, attr_value):
    """
    Set if component is active, used for read active attribute callback.
    Args:
        obj: object whose attribute is to be set
        attr_value: attribute value
    Returns:
        None
    """

    def change_active_attr(_obj, active):
        if hasattr(_obj, "_active"):
            _obj._active = active

    change_active_attr(obj, attr_value)
    if hasattr(obj, "_component") and obj._component is not None:
        change_active_attr(obj._component, attr_value)


def _set_sense(obj, attr_value):
    """
    Set if component is minimize or maximize.
    Args:
        obj: object whose attribute is to be set
        attr_value: attribute value
    Returns:
        None
    """
    obj._sense = attr_value
    obj._init_sense = attr_value


def _set_fixed(obj, attr_value):
    """
    Set if variable is fixed, used for read fixed attribute callback.
    Args:
        obj: object whose attribute is to be set
        attr_value: attribute value
    Returns:
        None
    """
    if attr_value:
        obj.fix()
    else:
        obj.unfix()


def _get_value(obj):
    """
    Get object value attribute callback.
    Args:
        obj: object whose attribute is to be set
    Returns:
        value
    """
    return value(obj, exception=False)


def _get_strict(obj):
    """
    Get object strict attribute callback.
    Args:
        obj: object
    Returns:
        strict value
    """
    return obj._strict


def _get_domain(obj):
    """
    Get object domain attribute callback.
    Args:
        obj: object
    Returns:
        domain string value
    """
    return obj.domain.name


def _get_index_name(obj: IndexedComponent) -> str:
    return obj.index_set().name


def _set_value(obj, attr_value):
    """
    Set object value attribute callback. We change even not  mutable objects.
    Args:
        obj: object whose attribute is to be set
        attr_value: attribute value
    Returns:
        None
    """
    try:
        obj.value = attr_value
    except AttributeError:
        pass


def _set_lb(obj, attr_value):
    """
    Set variable lower bound, used for read lb attribute callback.
    Args:
        obj: object whose attribute is to be set
        attr_value: attribute value
    Returns:
        None
    """
    obj.setlb(attr_value)


def _set_ub(obj, attr_value):
    """
    Set variable upper bound, use for read ub attribute callback.
    Args:
        obj: object whose attribute is to be set
        attr_value: attribute value
    Returns:
        None
    """
    obj.setub(attr_value)


def _set_strict(obj, attr_value):
    """
    Set variable strict attribute.
    Args:
        obj: object whose attribute is to be set
        attr_value: attribute value
    Returns:
        None
    """
    obj._strict = attr_value


def _only_fixed(obj, attrs):
    """
    Returns a list of attributes to read for a variable, only whether it is
    fixed for non-fixed variables and if it is fixed and the value for fixed
    variables.  The allows you to set up a serializer that only reads fixed
    variable values.
    Args:
        obj: Pyomo component being loaded
        attrs: State dictionary for the component obj.
    Returns:
        An attribute list to read. Loads fixed for either fixed or un-fixed
        variables, but only reads in values for unfixed variables.  This is
        useful for initialization functions.
    """
    if attrs["fixed"]:
        return "value", "fixed"
    else:
        return ("fixed",)


class StoreSpec(object):
    """
    A StoreSpec object tells the serializer functions what to read or write.
    The default settings will produce a StoreSpec configured to load/save the
    typical attributes required to load/save a model state.
    Args:
        classes: A list of classes to save.  Each class is represented by a
            list (or tuple) containing the following elements: (1) class
            (compared using isinstance) (2) attribute list or None, an empty
            list store the object, but none of its attributes, None will not
            store objects of this class type (3) optional load filter function.
            The load filter function returns a list of attributes to read based
            on the state of an object and its saved state. The allows, for
            example, loading values for unfixed variables, or only loading
            values whose current value is less than one. The filter function
            only applies to load not save. Filter functions take two arguments
            (a) the object (current state) and (b) the dictionary containing the
            saved state of an object.  More specific classes should come before
            more general classes.  For example if an object is a HeatExchanger
            and a UnitModel, and HeatExchanger is listed first, it will follow
            the HeatExchanger settings.  If UnitModel is listed first in the
            classes list, it will follow the UnitModel settings.
        data_classes: This takes the same form as the classes argument.
            This is for component data classes.
        skip_classes: This is a list of classes to skip.  If a class appears
            in the skip list, but also appears in the classes argument, the
            classes argument will override skip_classes. The use for this is to
            specifically exclude certain classes that would get caught by more
            general classes (e.g. UnitModel is in the class list, but you want
            to exclude HeatExchanger which is derived from UnitModel).
        ignore_missing: If True will ignore a component or attribute that exists
            in the model, but not in the stored state. If false an exception
            will be raised for things in the model that should be loaded but
            aren't in the stored state. Extra items in the stored state will not
            raise an exception regardless of this argument.
        suffix: If True store suffixes and component ids.  If false, don't store
            suffixes.
        suffix_filter: None to store all suffixes if suffix=True, or a list of
            suffixes to store if suffix=True
    """

    def __init__(
        self,
        classes=(
            (Param, ("_mutable",)),
            (Var, ()),
            (Expression, ()),
            (Component, ("active",)),
            (pyomo.core.base.objective.Objective, ("sense",)),
            (pyomo.core.base.indexed_component.IndexedComponent, ("index",)),
        ),
        data_classes=(
            (
                pyomo.core.base.var._VarData,
                ("fixed", "stale", "value", "lb", "ub", "domain"),
            ),
            (pyomo.core.base.param._ParamData, ("value",)),
            (int, ("value",)),
            (float, ("value",)),
            (pyomo.core.base.expression._ExpressionData, ()),
            (pyomo.core.base.component.ComponentData, ("active",)),
            (pyomo.core.base.constraint._GeneralConstraintData, ()),
            (pyomo.core.expr.numvalue.NumericConstant, ("value",)),
            (pyomo.core.expr.logical_expr.InequalityExpression, ("strict",)),
            (pyomo.core.base.objective.ScalarObjective, ("sense",)),
            (pyomo.core.base.set.RangeSet, ("_init_data",)),
        ),
        skip_classes=(ExternalFunction, Port, Expression, RangeSet),
        ignore_missing=True,
        suffix=True,
        suffix_filter=None,
    ):
        """
        (see above)
        """
        # Callbacks are used for attributes that cannot be directly get or set
        self.get_functions = {
            "value": _get_value,
            "strict": _get_strict,
            "domain": _get_domain,
            "index": _get_index_name,
        }
        self.set_functions = {
            "_mutable": lambda *args: None,
            "active": _set_active,
            "fixed": _set_fixed,
            "lb": _set_lb,
            "ub": _set_ub,
            "value": _set_value,
            "strict": _set_strict,
            "sense": _set_sense,
        }

        skip_with_classes = [(i, []) for i in skip_classes if i not in classes] + list(
            classes
        )
        self.classes = [i[0] for i in skip_with_classes]
        # Add skip classes to classes list, with None as attr list to skip
        self.class_attrs = [i[1] for i in skip_with_classes]
        self.data_classes = [i[0] for i in data_classes]
        self.data_class_attrs = [i[1] for i in data_classes]
        # Create filter function lists, use None if not supplied
        self.class_filter = [i[2] if len(i) > 2 else None for i in skip_with_classes]
        self.data_class_filter = [i[2] if len(i) > 2 else None for i in data_classes]
        self.ignore_missing = ignore_missing
        self.include_suffix = suffix
        self.suffix_filter = suffix_filter

    def set_read_callback(self, attr, cb=None):
        """
        Set a callback to set an attribute, when reading from json or dict.
        """
        self.set_functions[attr] = cb

    def set_write_callback(self, attr, cb=None):
        """
        Set a callback to get an attribute, when writing to json or dict.
        """
        self.get_functions[attr] = cb

    def get_class_attr_list(self, obj):
        """
        Look up what attributes to save/load for an Component object.
        Args:
            obj: Object to look up attribute list for.
        Return:
            A list of attributes and a filter function for object type
        """
        attr_list = []  # Attributes to store
        filter_function = None  # Load filter function
        for i, cl in enumerate(self.classes):
            if isinstance(obj, cl) or (isinstance(obj, type) and issubclass(obj, cl)):
                attr_list += list(self.class_attrs[i])
                filter_function = self.class_filter[i]  # this does not make sense
        return attr_list, filter_function

    def get_data_class_attr_list(self, obj):
        """
        Look up what attributes to save/load for an ComponentData object.
        Args:
            obj: Object or type to look up attribute list for.
        Return:
            A list of attributes and a filter function for object type
        """
        attr_list = []  # Attributes to store
        filter_function = None  # Load filter function
        for i, cl in enumerate(self.data_classes):
            if isinstance(obj, cl) or (isinstance(obj, type) and issubclass(obj, cl)):
                attr_list += list(self.data_class_attrs[i])
                filter_function = self.data_class_filter[i]  # this does not make sense
        return attr_list, filter_function

    @classmethod
    def bound(cls):
        """Returns a StoreSpec object to store variable bounds only."""
        return cls(
            classes=((Var, ()),),
            data_classes=((pyomo.core.base.var._VarData, ("lb", "ub")),),
            suffix=False,
        )

    @classmethod
    def value(cls):
        """Returns a StoreSpec object to store variable values only."""
        return cls(
            classes=((Var, ()),),
            data_classes=((pyomo.core.base.var._VarData, ("value",)),),
            suffix=False,
        )

    @classmethod
    def isfixed(cls):
        """Returns a StoreSpec object to store if variables are fixed."""
        return cls(
            classes=((Var, ()),),
            data_classes=((pyomo.core.base.var._VarData, ("fixed",)),),
            suffix=False,
        )

    @classmethod
    def suffix(cls, suffix_filter=None):
        return cls(
            classes=((Suffix, ()),),
            data_classes=(),
            suffix=True,
            suffix_filter=suffix_filter,
        )

    @classmethod
    def value_isfixed(cls, only_fixed):
        """
        Return a StoreSpec object to store variable values and if fixed.
        Args:
            only_fixed: Only load fixed variable values
        """
        if only_fixed:
            return cls(
                classes=((Var, ()),),
                data_classes=(
                    (pyomo.core.base.var._VarData, ("value", "fixed"), _only_fixed),
                ),
                suffix=False,
            )
        else:
            return cls(
                classes=((Var, ()),),
                data_classes=((pyomo.core.base.var._VarData, ("value", "fixed")),),
                suffix=False,
            )

    @classmethod
    def value_isfixed_isactive(cls, only_fixed):
        """
        Return a StoreSpec object to store variable values, if variables are
        fixed and if components are active.
        Args:
            only_fixed: Only load fixed variable values
        """
        if only_fixed:
            return cls(
                classes=((Var, ()), (Param, ()), (Component, ("active",))),
                data_classes=(
                    (pyomo.core.base.var._VarData, ("value", "fixed"), _only_fixed),
                    (pyomo.core.base.param._ParamData, ("value",)),
                    (pyomo.core.base.component.ComponentData, ("active",)),
                ),
                suffix=False,
            )
        else:
            return cls(
                classes=((Var, ()), (Param, ()), (Component, ("active",))),
                data_classes=(
                    (pyomo.core.base.var._VarData, ("value", "fixed")),
                    (pyomo.core.base.param._ParamData, ("value",)),
                    (pyomo.core.base.component.ComponentData, ("active",)),
                ),
                suffix=False,
            )
